# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from pytest import mark
from unittest.mock import MagicMock
from supertokens_python import InputAppInfo, SupertokensConfig, init
from supertokens_python.normalised_url_domain import NormalisedURLDomain
from supertokens_python.normalised_url_path import NormalisedURLPath
from supertokens_python.recipe import session
from supertokens_python.recipe.session import SessionRecipe
from supertokens_python.recipe.session.asyncio import create_new_session

from tests.utils import clean_st, reset, setup_st, start_st


def setup_function(_):
    reset()
    clean_st()
    setup_st()


def teardown_function(_):
    reset()
    clean_st()


def testing_URL_path_normalisation():
    def normalise_url_path_or_throw_error(
        input: str,
    ):  # pylint: disable=redefined-builtin
        return NormalisedURLPath(input).get_as_string_dangerous()

    assert (
        normalise_url_path_or_throw_error("exists?email=john.doe%40gmail.com")
        == "/exists"
    )
    assert (
        normalise_url_path_or_throw_error(
            "/auth/email/exists?email=john.doe%40gmail.com"
        )
        == "/auth/email/exists"
    )
    assert normalise_url_path_or_throw_error("exists") == "/exists"
    assert normalise_url_path_or_throw_error("/exists") == "/exists"
    assert (
        normalise_url_path_or_throw_error("/exists?email=john.doe%40gmail.com")
        == "/exists"
    )
    assert normalise_url_path_or_throw_error("http://api.example.com") == ""
    assert normalise_url_path_or_throw_error("https://api.example.com") == ""
    assert normalise_url_path_or_throw_error("http://api.example.com?hello=1") == ""
    assert normalise_url_path_or_throw_error("http://api.example.com/hello") == "/hello"
    assert normalise_url_path_or_throw_error("http://api.example.com/") == ""
    assert normalise_url_path_or_throw_error("http://api.example.com:8080") == ""
    assert normalise_url_path_or_throw_error("api.example.com/") == ""
    assert normalise_url_path_or_throw_error("api.example.com#random") == ""
    assert normalise_url_path_or_throw_error(".example.com") == ""
    assert normalise_url_path_or_throw_error("api.example.com/?hello=1&bye=2") == ""

    assert (
        normalise_url_path_or_throw_error("http://api.example.com/one/two")
        == "/one/two"
    )
    assert normalise_url_path_or_throw_error("http://1.2.3.4/one/two") == "/one/two"
    assert normalise_url_path_or_throw_error("1.2.3.4/one/two") == "/one/two"
    assert (
        normalise_url_path_or_throw_error("https://api.example.com/one/two/")
        == "/one/two"
    )
    assert (
        normalise_url_path_or_throw_error("http://api.example.com/one/two?hello=1")
        == "/one/two"
    )
    assert (
        normalise_url_path_or_throw_error("http://api.example.com/hello/") == "/hello"
    )
    assert (
        normalise_url_path_or_throw_error("http://api.example.com/one/two/")
        == "/one/two"
    )
    assert (
        normalise_url_path_or_throw_error("http://api.example.com/one/two#random2")
        == "/one/two"
    )
    assert normalise_url_path_or_throw_error("api.example.com/one/two") == "/one/two"
    assert normalise_url_path_or_throw_error(".example.com/one/two") == "/one/two"
    assert (
        normalise_url_path_or_throw_error("api.example.com/one/two?hello=1&bye=2")
        == "/one/two"
    )

    assert normalise_url_path_or_throw_error("/one/two") == "/one/two"
    assert normalise_url_path_or_throw_error("one/two") == "/one/two"
    assert normalise_url_path_or_throw_error("one/two/") == "/one/two"
    assert normalise_url_path_or_throw_error("/one") == "/one"
    assert normalise_url_path_or_throw_error("one") == "/one"
    assert normalise_url_path_or_throw_error("one/") == "/one"
    assert normalise_url_path_or_throw_error("/one/two/") == "/one/two"
    assert normalise_url_path_or_throw_error("/one/two?hello=1") == "/one/two"
    assert normalise_url_path_or_throw_error("one/two?hello=1") == "/one/two"
    assert normalise_url_path_or_throw_error("/one/two/#randm,") == "/one/two"
    assert normalise_url_path_or_throw_error("one/two#random") == "/one/two"

    assert normalise_url_path_or_throw_error("localhost:4000/one/two") == "/one/two"
    assert normalise_url_path_or_throw_error("127.0.0.1:4000/one/two") == "/one/two"
    assert normalise_url_path_or_throw_error("127.0.0.1/one/two") == "/one/two"
    assert (
        normalise_url_path_or_throw_error("https://127.0.0.1:80/one/two") == "/one/two"
    )
    assert normalise_url_path_or_throw_error("/") == ""
    assert normalise_url_path_or_throw_error("") == ""

    assert (
        normalise_url_path_or_throw_error("/.netlify/functions/api")
        == "/.netlify/functions/api"
    )
    assert (
        normalise_url_path_or_throw_error("/netlify/.functions/api")
        == "/netlify/.functions/api"
    )
    assert (
        normalise_url_path_or_throw_error("app.example.com/.netlify/functions/api")
        == "/.netlify/functions/api"
    )
    assert (
        normalise_url_path_or_throw_error("app.example.com/netlify/.functions/api")
        == "/netlify/.functions/api"
    )
    assert normalise_url_path_or_throw_error("/app.example.com") == "/app.example.com"


def testing_URL_domain_normalisation():
    def normalise_url_domain_or_throw_error(
        input: str,
    ):  # pylint: disable=redefined-builtin
        return NormalisedURLDomain(input).get_as_string_dangerous()

    assert (
        normalise_url_domain_or_throw_error("http://api.example.com")
        == "http://api.example.com"
    )
    assert (
        normalise_url_domain_or_throw_error("https://api.example.com")
        == "https://api.example.com"
    )
    assert (
        normalise_url_domain_or_throw_error("http://api.example.com?hello=1")
        == "http://api.example.com"
    )
    assert (
        normalise_url_domain_or_throw_error("http://api.example.com/hello")
        == "http://api.example.com"
    )
    assert (
        normalise_url_domain_or_throw_error("http://api.example.com/")
        == "http://api.example.com"
    )
    assert (
        normalise_url_domain_or_throw_error("http://api.example.com#random2")
        == "http://api.example.com"
    )
    assert (
        normalise_url_domain_or_throw_error("http://api.example.com:8080")
        == "http://api.example.com:8080"
    )
    assert (
        normalise_url_domain_or_throw_error("api.example.com/")
        == "https://api.example.com"
    )
    assert (
        normalise_url_domain_or_throw_error("api.example.com")
        == "https://api.example.com"
    )
    assert (
        normalise_url_domain_or_throw_error("api.example.com#random")
        == "https://api.example.com"
    )
    assert normalise_url_domain_or_throw_error(".example.com") == "https://example.com"
    assert (
        normalise_url_domain_or_throw_error("api.example.com/?hello=1&bye=2")
        == "https://api.example.com"
    )
    assert normalise_url_domain_or_throw_error("localhost") == "http://localhost"
    assert (
        normalise_url_domain_or_throw_error("https://localhost") == "https://localhost"
    )

    assert (
        normalise_url_domain_or_throw_error("http://api.example.com/one/two")
        == "http://api.example.com"
    )
    assert (
        normalise_url_domain_or_throw_error("http://1.2.3.4/one/two")
        == "http://1.2.3.4"
    )
    assert (
        normalise_url_domain_or_throw_error("https://1.2.3.4/one/two")
        == "https://1.2.3.4"
    )
    assert normalise_url_domain_or_throw_error("1.2.3.4/one/two") == "http://1.2.3.4"
    assert (
        normalise_url_domain_or_throw_error("https://api.example.com/one/two/")
        == "https://api.example.com"
    )
    assert (
        normalise_url_domain_or_throw_error("http://api.example.com/one/two?hello=1")
        == "http://api.example.com"
    )
    assert (
        normalise_url_domain_or_throw_error("http://api.example.com/one/two#random2")
        == "http://api.example.com"
    )
    assert (
        normalise_url_domain_or_throw_error("api.example.com/one/two")
        == "https://api.example.com"
    )
    assert (
        normalise_url_domain_or_throw_error(".example.com/one/two")
        == "https://example.com"
    )
    assert (
        normalise_url_domain_or_throw_error("localhost:4000") == "http://localhost:4000"
    )
    assert (
        normalise_url_domain_or_throw_error("127.0.0.1:4000") == "http://127.0.0.1:4000"
    )
    assert normalise_url_domain_or_throw_error("127.0.0.1") == "http://127.0.0.1"
    assert (
        normalise_url_domain_or_throw_error("https://127.0.0.1:80/")
        == "https://127.0.0.1:80"
    )

    try:
        normalise_url_domain_or_throw_error("/one/two")
    except Exception as e:
        assert str(e) == "Please provide a valid domain name"

    try:
        normalise_url_domain_or_throw_error("/.netlify/functions/api")
    except Exception as e:
        assert str(e) == "Please provide a valid domain name"


@mark.asyncio
async def test_same_site_values():
    start_st()

    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="api.supertokens.io",
            website_domain="supertokens.io",
        ),
        framework="fastapi",
        recipe_list=[session.init(cookie_same_site="lax")],
    )

    assert SessionRecipe.get_instance().config.cookie_same_site == "lax"

    reset()

    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="api.supertokens.io",
            website_domain="supertokens.io",
        ),
        framework="fastapi",
        recipe_list=[session.init(cookie_same_site="none")],
    )

    assert SessionRecipe.get_instance().config.cookie_same_site == "none"

    reset()

    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="api.supertokens.io",
            website_domain="supertokens.io",
        ),
        framework="fastapi",
        recipe_list=[session.init(cookie_same_site="strict")],
    )

    assert SessionRecipe.get_instance().config.cookie_same_site == "strict"

    reset()

    test_passed = True
    try:
        init(
            supertokens_config=SupertokensConfig("http://localhost:3567"),
            app_info=InputAppInfo(
                app_name="SuperTokens Demo",
                api_domain="api.supertokens.io",
                website_domain="supertokens.io",
            ),
            framework="fastapi",
            recipe_list=[session.init(cookie_same_site="random")],  # type: ignore
        )
        test_passed = False
    except Exception as e:
        assert str(e) == 'cookie same site must be one of "strict", "lax", or "none"'

    assert test_passed
    reset()

    test_passed = True
    try:
        init(
            supertokens_config=SupertokensConfig("http://localhost:3567"),
            app_info=InputAppInfo(
                app_name="SuperTokens Demo",
                api_domain="api.supertokens.io",
                website_domain="supertokens.io",
            ),
            framework="fastapi",
            recipe_list=[session.init(cookie_same_site=" ")],  # type: ignore
        )
        test_passed = False
    except Exception as e:
        assert str(e) == 'cookie same site must be one of "strict", "lax", or "none"'

    assert test_passed
    reset()

    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="api.supertokens.io",
            website_domain="supertokens.io",
        ),
        framework="fastapi",
        recipe_list=[
            session.init(get_token_transfer_method=lambda _, __, ___: "cookie")
        ],
    )

    assert SessionRecipe.get_instance().config.cookie_same_site == "lax"

    reset()

    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="https://platform-services-uat.com",
            website_domain="https://platform-ui-uat.com",
        ),
        framework="fastapi",
        recipe_list=[
            session.init(get_token_transfer_method=lambda _, __, ___: "cookie")
        ],
    )

    assert SessionRecipe.get_instance().config.cookie_same_site == "none"

    reset()


@mark.asyncio
async def test_config_values():
    start_st()

    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="api.supertokens.io",
            website_domain="supertokens.io",
            api_base_path="/",
        ),
        framework="fastapi",
        recipe_list=[session.init(anti_csrf="VIA_CUSTOM_HEADER")],
    )

    assert SessionRecipe.get_instance().config.cookie_same_site == "lax"
    assert SessionRecipe.get_instance().config.anti_csrf == "VIA_CUSTOM_HEADER"
    assert SessionRecipe.get_instance().config.cookie_secure

    reset()

    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="https://api.supertokens.io",
            website_domain="supertokens.io",
            api_base_path="test/",
            website_base_path="test1/",
        ),
        framework="fastapi",
        recipe_list=[
            session.init(get_token_transfer_method=lambda _, __, ___: "cookie")
        ],
    )

    assert SessionRecipe.get_instance().config.cookie_same_site == "lax"
    assert SessionRecipe.get_instance().config.anti_csrf == "NONE"
    assert SessionRecipe.get_instance().config.cookie_secure

    reset()

    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="api.supertokens.com",
            website_domain="supertokens.io",
            api_base_path="test/",
            website_base_path="test1/",
        ),
        framework="fastapi",
        recipe_list=[
            session.init(get_token_transfer_method=lambda _, __, ___: "cookie")
        ],
    )

    assert SessionRecipe.get_instance().config.cookie_same_site == "none"
    assert SessionRecipe.get_instance().config.anti_csrf == "VIA_CUSTOM_HEADER"
    assert SessionRecipe.get_instance().config.cookie_secure

    reset()

    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="api.supertokens.co.uk",
            website_domain="supertokens.co.uk",
            api_base_path="test/",
            website_base_path="test1/",
        ),
        framework="fastapi",
        recipe_list=[
            session.init(get_token_transfer_method=lambda _, __, ___: "cookie")
        ],
    )

    assert SessionRecipe.get_instance().config.cookie_same_site == "lax"
    assert SessionRecipe.get_instance().config.anti_csrf == "NONE"
    assert SessionRecipe.get_instance().config.cookie_secure

    reset()

    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="127.0.0.1:3000",
            website_domain="127.0.0.1:9000",
            api_base_path="test/",
            website_base_path="test1/",
        ),
        framework="fastapi",
        recipe_list=[
            session.init(get_token_transfer_method=lambda _, __, ___: "cookie")
        ],
    )

    assert SessionRecipe.get_instance().config.cookie_same_site == "lax"
    assert SessionRecipe.get_instance().config.anti_csrf == "NONE"
    assert not SessionRecipe.get_instance().config.cookie_secure

    reset()

    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="127.0.0.1:3000",
            website_domain="127.0.0.1:9000",
            api_base_path="test/",
            website_base_path="test1/",
        ),
        framework="fastapi",
        recipe_list=[session.init(anti_csrf="VIA_CUSTOM_HEADER")],
    )

    assert SessionRecipe.get_instance().config.cookie_same_site == "lax"
    assert SessionRecipe.get_instance().config.anti_csrf == "VIA_CUSTOM_HEADER"
    assert not SessionRecipe.get_instance().config.cookie_secure

    reset()

    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="api.supertokens.io",
            website_domain="127.0.0.1:9000",
            api_base_path="test/",
            website_base_path="test1/",
        ),
        framework="fastapi",
        recipe_list=[
            session.init(get_token_transfer_method=lambda _, __, ___: "cookie")
        ],
    )

    assert SessionRecipe.get_instance().config.cookie_same_site == "none"
    assert SessionRecipe.get_instance().config.anti_csrf == "VIA_CUSTOM_HEADER"
    assert SessionRecipe.get_instance().config.cookie_secure

    reset()

    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="127.0.0.1:3000",
            website_domain="127.0.0.1:9000",
            api_base_path="test/",
            website_base_path="test1/",
        ),
        framework="fastapi",
        recipe_list=[
            session.init(get_token_transfer_method=lambda _, __, ___: "cookie")
        ],
    )

    assert SessionRecipe.get_instance().config.cookie_same_site == "lax"
    assert SessionRecipe.get_instance().config.anti_csrf == "NONE"
    assert not SessionRecipe.get_instance().config.cookie_secure

    reset()

    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="https://localhost",
            website_domain="http://localhost:3000",
            api_base_path="test/",
            website_base_path="test1/",
        ),
        framework="fastapi",
        recipe_list=[
            session.init(get_token_transfer_method=lambda _, __, ___: "cookie")
        ],
    )

    assert SessionRecipe.get_instance().config.cookie_same_site == "none"
    assert SessionRecipe.get_instance().config.cookie_secure

    reset()


def testing_override_test():
    m = None

    class OI:
        def __init__(self):
            pass

        def some_other_func(self):  # pylint: disable=no-self-use
            nonlocal m
            m = 1

        def some_func(self):
            self.some_other_func()

    class A(OI):
        def some_other_func(self):
            nonlocal m
            m = 2

    b = A()

    def some_other_func():
        nonlocal m
        m = 4

    b.some_other_func = some_other_func

    b.some_func()

    assert m == 4


def testing_super_recipe_tests():
    m = 0

    class EP:
        def __init__(self):
            pass

        def sign_up(self):
            self.get_user()

        def get_user(self):  # pylint: disable=no-self-use
            nonlocal m
            m = 1

    class TPEP:
        def __init__(self):
            ep = EP()
            self.o_sign_up = ep.sign_up
            self.o_get_user = ep.get_user
            dep = DerivedEP(self)
            ep.sign_up = dep.sign_up
            ep.get_user = dep.get_user

        def sign_up(self):
            self.o_sign_up()

        def get_users(self):
            self.o_get_user()

    class DerivedEP(EP):
        def __init__(self, tpep: TPEP):
            super().__init__()
            self.tpep = tpep

        def sign_up(self):
            self.tpep.sign_up()

        def get_user(self):
            self.tpep.get_users()

    def override(tpep: TPEP):
        o_sign_up = tpep.sign_up
        o_get_users = tpep.get_users

        def sign_up():
            o_sign_up()

        def get_users():
            nonlocal m
            m = 5
            o_get_users()
            if m == 1:  # type: ignore
                m = 2

        tpep.sign_up = sign_up
        tpep.get_users = get_users
        return tpep

    base_tpep = TPEP()

    override_tpep = override(base_tpep)

    override_tpep.sign_up()
    assert m == 2

    m = 1
    ep = DerivedEP(override_tpep)

    ep.get_user()

    assert m == 2


@mark.asyncio
async def test_samesite_valid_config():
    domain_combinations = [
        ["http://localhost:3000", "http://localhost:8000"],
        ["http://127.0.0.1:3000", "http://localhost:8000"],
        ["http://localhost:3000", "http://127.0.0.1:8000"],
        ["http://127.0.0.1:3000", "http://127.0.0.1:8000"],
        ["https://localhost:3000", "https://localhost:8000"],
        ["https://127.0.0.1:3000", "https://localhost:8000"],
        ["https://localhost:3000", "https://127.0.0.1:8000"],
        ["https://127.0.0.1:3000", "https://127.0.0.1:8000"],
        ["https://supertokens.io", "https://api.supertokens.io"],
        ["https://supertokens.io", "https://supertokensapi.io"],
        ["http://localhost:3000", "https://supertokensapi.io"],
        ["http://127.0.0.1:3000", "https://supertokensapi.io"],
    ]
    for (website_domain, api_domain) in domain_combinations:
        reset()
        clean_st()
        setup_st()

        init(
            supertokens_config=SupertokensConfig("http://localhost:3567"),
            app_info=InputAppInfo(
                app_name="SuperTokens Demo",
                website_domain=website_domain,
                api_domain=api_domain,
            ),
            framework="fastapi",
            recipe_list=[session.init(cookie_same_site="none")],
        )


@mark.asyncio
async def test_samesite_invalid_config():
    domain_combinations = [
        ["http://localhost:3000", "http://supertokensapi.io"],
        ["http://127.0.0.1:3000", "http://supertokensapi.io"],
        ["http://supertokens.io", "http://localhost:8000"],
        ["http://supertokens.io", "http://127.0.0.1:8000"],
        ["http://supertokens.io", "http://supertokensapi.io"],
    ]
    for (website_domain, api_domain) in domain_combinations:
        reset()
        clean_st()
        setup_st()
        try:
            init(
                supertokens_config=SupertokensConfig("http://localhost:3567"),
                app_info=InputAppInfo(
                    app_name="SuperTokens Demo",
                    website_domain=website_domain,
                    api_domain=api_domain,
                ),
                framework="fastapi",
                recipe_list=[
                    session.init(
                        cookie_same_site="none",
                        get_token_transfer_method=lambda _, __, ___: "cookie",
                    )
                ],
            )
            await create_new_session(MagicMock(), "userId", {}, {})
        except Exception as e:
            assert (
                str(e)
                == "Since your API and website domain are different, for sessions to work, please use https on your apiDomain and don't set cookieSecure to false."
            )
        else:
            assert False, "Exception not raised"



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Any, Callable, Dict, Optional, TypeVar, Union, Generic, List

from supertokens_python.types import MaybeAwaitable
from supertokens_python.utils import get_timestamp_ms

from ..interfaces import (
    JSONObject,
    JSONPrimitive,
    SessionClaim,
    SessionClaimValidator,
    ClaimValidationResult,
    JSONPrimitiveList,
)


Primitive = TypeVar("Primitive", bound=JSONPrimitive)
PrimitiveList = TypeVar("PrimitiveList", bound=JSONPrimitiveList)

_T = TypeVar("_T")


class SCVMixin(SessionClaimValidator, Generic[_T]):
    def __init__(
        self,
        id_: str,
        claim: SessionClaim[PrimitiveList],
        val: _T,
        max_age_in_sec: Optional[int] = None,
    ):
        super().__init__(id_)
        self.claim: SessionClaim[PrimitiveList] = claim  # TODO:PrimitiveArrayClaim
        self.val = val
        self.max_age_in_sec = max_age_in_sec

    def should_refetch(
        self,
        payload: JSONObject,
        user_context: Dict[str, Any],
    ) -> bool:
        claim = self.claim

        return (claim.get_value_from_payload(payload, user_context) is None) or (
            self.max_age_in_sec is not None
            and (
                payload[claim.key]["t"]
                < get_timestamp_ms() - self.max_age_in_sec * 1000
            )
        )

    async def _validate(
        self,
        payload: JSONObject,
        user_context: Dict[str, Any],
        is_include: bool,
    ):
        val = self.val
        max_age_in_sec = self.max_age_in_sec

        expected_key = "expectedToInclude" if is_include else "expectedToNotInclude"

        assert isinstance(self.claim, PrimitiveArrayClaim)
        claim_val = self.claim.get_value_from_payload(payload, user_context)

        if claim_val is None:
            return ClaimValidationResult(
                is_valid=False,
                reason={
                    "message": "value does not exist",
                    expected_key: val,
                    "actualValue": claim_val,
                },
            )

        last_refetch_time = self.claim.get_last_refetch_time(payload, user_context)
        assert last_refetch_time is not None
        age_in_sec = (get_timestamp_ms() - last_refetch_time) / 1000
        if max_age_in_sec is not None and age_in_sec > max_age_in_sec:
            return ClaimValidationResult(
                is_valid=False,
                reason={
                    "message": "expired",
                    "ageInSeconds": age_in_sec,
                    "maxAgeInSeconds": max_age_in_sec,
                },
            )

        # Doing this to ensure same code in the upcoming steps irrespective of
        # whether self.val is Primitive or PrimitiveList
        vals: List[JSONPrimitive] = (
            val if isinstance(val, list) else [val]
        )  # pyright: reportGeneralTypeIssues=false

        claim_val_set = set(claim_val)
        if is_include:
            for v in vals:
                if v not in claim_val_set:
                    return ClaimValidationResult(
                        is_valid=False,
                        reason={
                            "message": "wrong value",
                            expected_key: val,
                            # other SDKs return the item itself
                            "actualValue": claim_val,
                        },
                    )
        else:
            for v in vals:
                if v in claim_val_set:
                    return ClaimValidationResult(
                        is_valid=False,
                        reason={
                            "message": "wrong value",
                            expected_key: val,
                            # other SDKs return the item itself
                            "actualValue": claim_val,
                        },
                    )

        return ClaimValidationResult(is_valid=True)


class IncludesSCV(SCVMixin[Primitive]):
    async def validate(
        self,
        payload: JSONObject,
        user_context: Dict[str, Any],
    ):
        return await self._validate(payload, user_context, is_include=True)


class ExcludesSCV(SCVMixin[Primitive]):
    async def validate(
        self,
        payload: JSONObject,
        user_context: Dict[str, Any],
    ):
        return await self._validate(payload, user_context, is_include=False)


class IncludesAllSCV(SCVMixin[PrimitiveList]):
    async def validate(
        self,
        payload: JSONObject,
        user_context: Dict[str, Any],
    ):
        return await self._validate(payload, user_context, is_include=True)


class ExcludesAllSCV(SCVMixin[PrimitiveList]):
    async def validate(
        self,
        payload: JSONObject,
        user_context: Dict[str, Any],
    ):
        return await self._validate(payload, user_context, is_include=False)


class PrimitiveArrayClaimValidators(Generic[PrimitiveList]):
    def __init__(
        self,
        claim: SessionClaim[PrimitiveList],
        default_max_age_in_sec: Optional[int] = None,
    ) -> None:
        self.claim = claim
        self.default_max_age_in_sec = default_max_age_in_sec

    def includes(  # pyright: ignore[reportInvalidTypeVarUse]
        self,
        val: Primitive,  # pyright: ignore[reportInvalidTypeVarUse]
        max_age_in_seconds: Optional[int] = None,
        id_: Union[str, None] = None,
    ) -> SessionClaimValidator:
        max_age_in_sec = max_age_in_seconds or self.default_max_age_in_sec
        return IncludesSCV(
            (id_ or self.claim.key), self.claim, val=val, max_age_in_sec=max_age_in_sec
        )

    def excludes(  # pyright: ignore[reportInvalidTypeVarUse]
        self,
        val: Primitive,  # pyright: ignore[reportInvalidTypeVarUse]
        max_age_in_seconds: Optional[int] = None,
        id_: Union[str, None] = None,
    ) -> SessionClaimValidator:
        max_age_in_sec = max_age_in_seconds or self.default_max_age_in_sec
        return ExcludesSCV(
            (id_ or self.claim.key), self.claim, val=val, max_age_in_sec=max_age_in_sec
        )

    def includes_all(
        self,
        val: PrimitiveList,
        max_age_in_seconds: Optional[int] = None,
        id_: Union[str, None] = None,
    ) -> SessionClaimValidator:
        max_age_in_sec = max_age_in_seconds or self.default_max_age_in_sec
        return IncludesAllSCV(
            (id_ or self.claim.key), self.claim, val=val, max_age_in_sec=max_age_in_sec
        )

    def excludes_all(
        self,
        val: PrimitiveList,
        max_age_in_seconds: Optional[int] = None,
        id_: Union[str, None] = None,
    ) -> SessionClaimValidator:
        max_age_in_sec = max_age_in_seconds or self.default_max_age_in_sec
        return ExcludesAllSCV(
            (id_ or self.claim.key), self.claim, val=val, max_age_in_sec=max_age_in_sec
        )


class PrimitiveArrayClaim(SessionClaim[PrimitiveList], Generic[PrimitiveList]):
    def __init__(
        self,
        key: str,
        fetch_value: Callable[
            [str, Dict[str, Any]],
            MaybeAwaitable[Optional[PrimitiveList]],
        ],
        default_max_age_in_sec: Optional[int] = None,
    ) -> None:
        super().__init__(key, fetch_value)

        claim = self
        self.validators = PrimitiveArrayClaimValidators(claim, default_max_age_in_sec)

    def add_to_payload_(
        self,
        payload: Dict[str, Any],
        value: PrimitiveList,
        user_context: Union[Dict[str, Any], None] = None,
    ) -> JSONObject:
        payload[self.key] = {"v": value, "t": get_timestamp_ms()}
        _ = user_context

        return payload

    def remove_from_payload_by_merge_(
        self, payload: JSONObject, user_context: Optional[Dict[str, Any]] = None
    ) -> JSONObject:
        payload[self.key] = None
        return payload

    def remove_from_payload(
        self, payload: JSONObject, user_context: Optional[Dict[str, Any]] = None
    ) -> JSONObject:
        del payload[self.key]
        return payload

    def get_value_from_payload(
        self, payload: JSONObject, user_context: Union[Dict[str, Any], None] = None
    ) -> Union[PrimitiveList, None]:
        _ = user_context

        return payload.get(self.key, {}).get("v")

    def get_last_refetch_time(
        self, payload: JSONObject, user_context: Union[Dict[str, Any], None] = None
    ) -> Union[int, None]:
        _ = user_context

        return payload.get(self.key, {}).get("t")



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Dict, Union

from typing_extensions import Literal

if TYPE_CHECKING:
    from ...recipe_module import RecipeModule
    from supertokens_python.supertokens import AppInfo, BaseRequest
    from .utils import TokenTransferMethod

from . import exceptions as ex
from . import interfaces, utils
from .recipe import SessionRecipe

InputErrorHandlers = utils.InputErrorHandlers
InputOverrideConfig = utils.InputOverrideConfig
JWTConfig = utils.JWTConfig
SessionContainer = interfaces.SessionContainer
exceptions = ex


def init(
    cookie_domain: Union[str, None] = None,
    cookie_secure: Union[bool, None] = None,
    cookie_same_site: Union[Literal["lax", "none", "strict"], None] = None,
    session_expired_status_code: Union[int, None] = None,
    anti_csrf: Union[Literal["VIA_TOKEN", "VIA_CUSTOM_HEADER", "NONE"], None] = None,
    get_token_transfer_method: Union[
        Callable[
            [BaseRequest, bool, Dict[str, Any]],
            Union[TokenTransferMethod, Literal["any"]],
        ],
        None,
    ] = None,
    error_handlers: Union[InputErrorHandlers, None] = None,
    override: Union[InputOverrideConfig, None] = None,
    jwt: Union[JWTConfig, None] = None,
    invalid_claim_status_code: Union[int, None] = None,
) -> Callable[[AppInfo], RecipeModule]:
    return SessionRecipe.init(
        cookie_domain,
        cookie_secure,
        cookie_same_site,
        session_expired_status_code,
        anti_csrf,
        get_token_transfer_method,
        error_handlers,
        override,
        jwt,
        invalid_claim_status_code,
    )



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from typing import Union

import phonenumbers  # type: ignore
from phonenumbers import format_number, parse  # type: ignore
from supertokens_python.exceptions import raise_bad_input_exception
from supertokens_python.recipe.passwordless.interfaces import APIInterface, APIOptions
from supertokens_python.recipe.passwordless.utils import (
    ContactEmailOnlyConfig,
    ContactEmailOrPhoneConfig,
    ContactPhoneOnlyConfig,
)
from supertokens_python.types import GeneralErrorResponse
from supertokens_python.utils import default_user_context, send_200_response


async def create_code(api_implementation: APIInterface, api_options: APIOptions):
    if api_implementation.disable_create_code_post:
        return None

    body = await api_options.request.json()

    if body is None:
        raise_bad_input_exception("Please provide a JSON body")

    email: Union[str, None] = None
    phone_number: Union[str, None] = None

    if ("email" in body and "phoneNumber" in body) or (
        "email" not in body and "phoneNumber" not in body
    ):
        raise_bad_input_exception("Please provide exactly one of email or phoneNumber")

    if "email" not in body and isinstance(
        api_options.config.contact_config, ContactEmailOnlyConfig
    ):
        raise_bad_input_exception(
            'Please provide an email since you have set the contactMethod to "EMAIL"'
        )

    if "phoneNumber" not in body and isinstance(
        api_options.config.contact_config, ContactPhoneOnlyConfig
    ):
        raise_bad_input_exception(
            'Please provide a phoneNumber since you have set the contactMethod to "PHONE"'
        )

    if "email" in body:
        email = body["email"]
    if "phoneNumber" in body:
        phone_number = body["phoneNumber"]

    if email is not None and (
        isinstance(
            api_options.config.contact_config,
            (ContactEmailOnlyConfig, ContactEmailOrPhoneConfig),
        )
    ):
        email = email.strip()
        validation_error = (
            await api_options.config.contact_config.validate_email_address(email)
        )
        if validation_error is not None:
            api_options.response.set_json_content(
                GeneralErrorResponse(validation_error).to_json()
            )
            return api_options.response

    if phone_number is not None and (
        isinstance(
            api_options.config.contact_config,
            (ContactEmailOrPhoneConfig, ContactPhoneOnlyConfig),
        )
    ):
        validation_error = (
            await api_options.config.contact_config.validate_phone_number(phone_number)
        )
        if validation_error is not None:
            api_options.response.set_json_content(
                GeneralErrorResponse(validation_error).to_json()
            )
            return api_options.response
        try:
            phone_number_formatted: str = format_number(
                parse(phone_number, None), phonenumbers.PhoneNumberFormat.E164
            )  # type: ignore
            phone_number = phone_number_formatted
        except Exception:
            phone_number = phone_number.strip()

    user_context = default_user_context(api_options.request)

    result = await api_implementation.create_code_post(
        email=email,
        phone_number=phone_number,
        api_options=api_options,
        user_context=user_context,
    )
    return send_200_response(result.to_json(), api_options.response)



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from typing import Any, Dict, Callable, Union

from supertokens_python.ingredients.smsdelivery.types import (
    SMSDeliveryInterface,
    TwilioServiceInterface,
    TwilioSettings,
)
from supertokens_python.recipe.passwordless.smsdelivery.services.twilio import (
    TwilioService as PlessTwilioService,
)

from ....types import SMSTemplateVars


class TwilioService(SMSDeliveryInterface[SMSTemplateVars]):
    pless_twilio_service: PlessTwilioService

    def __init__(
        self,
        twilio_settings: TwilioSettings,
        override: Union[
            Callable[
                [TwilioServiceInterface[SMSTemplateVars]],
                TwilioServiceInterface[SMSTemplateVars],
            ],
            None,
        ] = None,
    ) -> None:
        self.pless_twilio_service = PlessTwilioService(twilio_settings, override)

    async def send_sms(
        self, template_vars: SMSTemplateVars, user_context: Dict[str, Any]
    ) -> None:
        await self.pless_twilio_service.send_sms(template_vars, user_context)



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Dict, List, Union

from httpx import AsyncClient
from supertokens_python.recipe.thirdparty.provider import Provider
from supertokens_python.recipe.thirdparty.types import (
    AccessTokenAPI,
    AuthorisationRedirectAPI,
    UserInfo,
    UserInfoEmail,
)

if TYPE_CHECKING:
    from supertokens_python.framework.request import BaseRequest

from supertokens_python.utils import get_filtered_list


class Github(Provider):
    def __init__(
        self,
        client_id: str,
        client_secret: str,
        scope: Union[None, List[str]] = None,
        authorisation_redirect: Union[
            None, Dict[str, Union[str, Callable[[BaseRequest], str]]]
        ] = None,
        is_default: bool = False,
    ):
        super().__init__("github", is_default)
        default_scopes = ["read:user", "user:email"]
        if scope is None:
            scope = default_scopes
        self.client_id = client_id
        self.client_secret = client_secret
        self.scopes = list(set(scope))
        self.access_token_api_url = "https://github.com/login/oauth/access_token"
        self.authorisation_redirect_url = "https://github.com/login/oauth/authorize"
        self.authorisation_redirect_params = {}
        if authorisation_redirect is not None:
            self.authorisation_redirect_params = authorisation_redirect

    async def get_profile_info(
        self, auth_code_response: Dict[str, Any], user_context: Dict[str, Any]
    ) -> UserInfo:
        access_token: str = auth_code_response["access_token"]
        params = {"alt": "json"}
        headers = {
            "Authorization": "Bearer " + access_token,
            "Accept": "application/vnd.github.v3+json",
        }
        async with AsyncClient() as client:
            response_user = await client.get(  # type:ignore
                url="https://api.github.com/user", params=params, headers=headers
            )
            response_email = await client.get(  # type:ignore
                url="https://api.github.com/user/emails", params=params, headers=headers
            )
            user_info = response_user.json()
            emails_info = response_email.json()
            user_id = str(user_info["id"])
            email_info = get_filtered_list(
                lambda x: "primary" in x and x["primary"], emails_info
            )

            if len(email_info) == 0:
                return UserInfo(user_id)
            is_email_verified = (
                email_info[0]["verified"] if "verified" in email_info[0] else False
            )
            email = (
                email_info[0]["email"]
                if "email" in email_info[0]
                else user_info["email"]
            )
            return UserInfo(user_id, UserInfoEmail(email, is_email_verified))

    def get_authorisation_redirect_api_info(
        self, user_context: Dict[str, Any]
    ) -> AuthorisationRedirectAPI:
        params = {
            "scope": " ".join(self.scopes),
            "client_id": self.client_id,
            **self.authorisation_redirect_params,
        }
        return AuthorisationRedirectAPI(self.authorisation_redirect_url, params)

    def get_access_token_api_info(
        self,
        redirect_uri: str,
        auth_code_from_request: str,
        user_context: Dict[str, Any],
    ) -> AccessTokenAPI:
        params = {
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "code": auth_code_from_request,
            "redirect_uri": redirect_uri,
        }
        return AccessTokenAPI(self.access_token_api_url, params)

    def get_redirect_uri(self, user_context: Dict[str, Any]) -> Union[None, str]:
        return None

    def get_client_id(self, user_context: Dict[str, Any]) -> str:
        return self.client_id



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from os import environ
from typing import TYPE_CHECKING, Any, Awaitable, Callable, Dict, List, Union

from supertokens_python.ingredients.emaildelivery import EmailDeliveryIngredient
from supertokens_python.ingredients.emaildelivery.types import EmailDeliveryConfig
from supertokens_python.ingredients.smsdelivery import SMSDeliveryIngredient
from supertokens_python.querier import Querier
from supertokens_python.recipe.passwordless.types import (
    PasswordlessIngredients,
    PasswordlessLoginSMSTemplateVars,
)
from typing_extensions import Literal

from .api import (
    consume_code,
    create_code,
    email_exists,
    phone_number_exists,
    resend_code,
)
from .api.implementation import APIImplementation
from .constants import (
    CONSUME_CODE_API,
    CREATE_CODE_API,
    DOES_EMAIL_EXIST_API,
    DOES_PHONE_NUMBER_EXIST_API,
    RESEND_CODE_API,
)
from .exceptions import SuperTokensPasswordlessError
from .interfaces import (
    APIOptions,
    ConsumeCodeOkResult,
    RecipeInterface,
    PasswordlessLoginEmailTemplateVars,
)
from .recipe_implementation import RecipeImplementation
from .utils import (
    ContactConfig,
    OverrideConfig,
    validate_and_normalise_user_input,
)
from ..emailverification import EmailVerificationRecipe
from ..emailverification.interfaces import (
    GetEmailForUserIdOkResult,
    EmailDoesNotExistError,
    UnknownUserIdError,
)
from ...post_init_callbacks import PostSTInitCallbacks

if TYPE_CHECKING:
    from supertokens_python.framework.request import BaseRequest
    from supertokens_python.framework.response import BaseResponse
    from supertokens_python.supertokens import AppInfo

from supertokens_python.exceptions import SuperTokensError, raise_general_exception
from supertokens_python.ingredients.smsdelivery.types import SMSDeliveryConfig
from supertokens_python.normalised_url_path import NormalisedURLPath
from supertokens_python.recipe_module import APIHandled, RecipeModule


class PasswordlessRecipe(RecipeModule):
    recipe_id = "passwordless"
    __instance = None
    email_delivery: EmailDeliveryIngredient[PasswordlessLoginEmailTemplateVars]
    sms_delivery: SMSDeliveryIngredient[PasswordlessLoginSMSTemplateVars]

    def __init__(
        self,
        recipe_id: str,
        app_info: AppInfo,
        contact_config: ContactConfig,
        flow_type: Literal[
            "USER_INPUT_CODE", "MAGIC_LINK", "USER_INPUT_CODE_AND_MAGIC_LINK"
        ],
        ingredients: PasswordlessIngredients,
        override: Union[OverrideConfig, None] = None,
        get_custom_user_input_code: Union[
            Callable[[Dict[str, Any]], Awaitable[str]], None
        ] = None,
        email_delivery: Union[
            EmailDeliveryConfig[PasswordlessLoginEmailTemplateVars], None
        ] = None,
        sms_delivery: Union[
            SMSDeliveryConfig[PasswordlessLoginSMSTemplateVars], None
        ] = None,
    ):
        super().__init__(recipe_id, app_info)
        self.config = validate_and_normalise_user_input(
            app_info,
            contact_config,
            flow_type,
            override,
            get_custom_user_input_code,
            email_delivery,
            sms_delivery,
        )

        recipe_implementation = RecipeImplementation(Querier.get_instance(recipe_id))
        self.recipe_implementation: RecipeInterface = (
            recipe_implementation
            if self.config.override.functions is None
            else self.config.override.functions(recipe_implementation)
        )

        api_implementation = APIImplementation()
        self.api_implementation = (
            api_implementation
            if self.config.override.apis is None
            else self.config.override.apis(api_implementation)
        )

        email_delivery_ingredient = ingredients.email_delivery
        if email_delivery_ingredient is None:
            self.email_delivery = EmailDeliveryIngredient(
                self.config.get_email_delivery_config()
            )
        else:
            self.email_delivery = email_delivery_ingredient

        sms_delivery_ingredient = ingredients.sms_delivery
        self.sms_delivery = (
            SMSDeliveryIngredient(self.config.get_sms_delivery_config())
            if sms_delivery_ingredient is None
            else sms_delivery_ingredient
        )

        def callback():
            ev_recipe = EmailVerificationRecipe.get_instance_optional()
            if ev_recipe:
                ev_recipe.add_get_email_for_user_id_func(self.get_email_for_user_id)

        PostSTInitCallbacks.add_post_init_callback(callback)

    def get_apis_handled(self) -> List[APIHandled]:
        return [
            APIHandled(
                method="post",
                path_without_api_base_path=NormalisedURLPath(CONSUME_CODE_API),
                request_id=CONSUME_CODE_API,
                disabled=self.api_implementation.disable_consume_code_post,
            ),
            APIHandled(
                method="post",
                path_without_api_base_path=NormalisedURLPath(CREATE_CODE_API),
                request_id=CREATE_CODE_API,
                disabled=self.api_implementation.disable_create_code_post,
            ),
            APIHandled(
                method="get",
                path_without_api_base_path=NormalisedURLPath(DOES_EMAIL_EXIST_API),
                request_id=DOES_EMAIL_EXIST_API,
                disabled=self.api_implementation.disable_email_exists_get,
            ),
            APIHandled(
                method="get",
                path_without_api_base_path=NormalisedURLPath(
                    DOES_PHONE_NUMBER_EXIST_API
                ),
                request_id=DOES_PHONE_NUMBER_EXIST_API,
                disabled=self.api_implementation.disable_phone_number_exists_get,
            ),
            APIHandled(
                method="post",
                path_without_api_base_path=NormalisedURLPath(RESEND_CODE_API),
                request_id=RESEND_CODE_API,
                disabled=self.api_implementation.disable_resend_code_post,
            ),
        ]

    async def handle_api_request(
        self,
        request_id: str,
        request: BaseRequest,
        path: NormalisedURLPath,
        method: str,
        response: BaseResponse,
    ):
        options = APIOptions(
            request,
            response,
            self.get_recipe_id(),
            self.config,
            self.recipe_implementation,
            self.get_app_info(),
            self.email_delivery,
            self.sms_delivery,
        )
        if request_id == CONSUME_CODE_API:
            return await consume_code(self.api_implementation, options)
        if request_id == CREATE_CODE_API:
            return await create_code(self.api_implementation, options)
        if request_id == DOES_EMAIL_EXIST_API:
            return await email_exists(self.api_implementation, options)
        if request_id == DOES_PHONE_NUMBER_EXIST_API:
            return await phone_number_exists(self.api_implementation, options)
        return await resend_code(self.api_implementation, options)

    async def handle_error(
        self, request: BaseRequest, err: SuperTokensError, response: BaseResponse
    ) -> BaseResponse:  # type: ignore
        raise err

    def get_all_cors_headers(self) -> List[str]:
        return []

    def is_error_from_this_recipe_based_on_instance(self, err: Exception) -> bool:
        return isinstance(err, SuperTokensError) and isinstance(
            err, SuperTokensPasswordlessError
        )

    @staticmethod
    def init(
        contact_config: ContactConfig,
        flow_type: Literal[
            "USER_INPUT_CODE", "MAGIC_LINK", "USER_INPUT_CODE_AND_MAGIC_LINK"
        ],
        override: Union[OverrideConfig, None] = None,
        get_custom_user_input_code: Union[
            Callable[[Dict[str, Any]], Awaitable[str]], None
        ] = None,
        email_delivery: Union[
            EmailDeliveryConfig[PasswordlessLoginEmailTemplateVars], None
        ] = None,
        sms_delivery: Union[
            SMSDeliveryConfig[PasswordlessLoginSMSTemplateVars], None
        ] = None,
    ):
        def func(app_info: AppInfo):
            if PasswordlessRecipe.__instance is None:
                ingredients = PasswordlessIngredients(None, None)
                PasswordlessRecipe.__instance = PasswordlessRecipe(
                    PasswordlessRecipe.recipe_id,
                    app_info,
                    contact_config,
                    flow_type,
                    ingredients,
                    override,
                    get_custom_user_input_code,
                    email_delivery,
                    sms_delivery,
                )
                return PasswordlessRecipe.__instance
            raise_general_exception(
                "Passwordless recipe has already been initialised. Please check "
                "your code for bugs."
            )

        return func

    @staticmethod
    def get_instance() -> PasswordlessRecipe:
        if PasswordlessRecipe.__instance is not None:
            return PasswordlessRecipe.__instance
        raise_general_exception(
            "Initialisation not done. Did you forget to call the SuperTokens.init function?"
        )

    @staticmethod
    def reset():
        if ("SUPERTOKENS_ENV" not in environ) or (
            environ["SUPERTOKENS_ENV"] != "testing"
        ):
            raise_general_exception("calling testing function in non testing env")
        PasswordlessRecipe.__instance = None

    async def create_magic_link(
        self,
        email: Union[str, None],
        phone_number: Union[str, None],
        user_context: Dict[str, Any],
    ) -> str:
        user_input_code = None
        if self.config.get_custom_user_input_code is not None:
            user_input_code = await self.config.get_custom_user_input_code(user_context)

        code_info = await self.recipe_implementation.create_code(
            email=email,
            phone_number=phone_number,
            user_input_code=user_input_code,
            user_context=user_context,
        )

        app_info = self.get_app_info()

        magic_link = (
            app_info.website_domain.get_as_string_dangerous()
            + app_info.website_base_path.get_as_string_dangerous()
            + "/verify"
            + "?rid="
            + self.get_recipe_id()
            + "&preAuthSessionId="
            + code_info.pre_auth_session_id
            + "#"
            + code_info.link_code
        )
        return magic_link

    async def signinup(
        self,
        email: Union[str, None],
        phone_number: Union[str, None],
        user_context: Dict[str, Any],
    ) -> ConsumeCodeOkResult:
        code_info = await self.recipe_implementation.create_code(
            email=email,
            phone_number=phone_number,
            user_context=user_context,
            user_input_code=None,
        )
        consume_code_result = await self.recipe_implementation.consume_code(
            link_code=code_info.link_code,
            pre_auth_session_id=code_info.pre_auth_session_id,
            device_id=code_info.device_id,
            user_input_code=code_info.user_input_code,
            user_context=user_context,
        )
        if isinstance(consume_code_result, ConsumeCodeOkResult):
            return consume_code_result
        raise Exception("Failed to create user. Please retry")

    async def get_email_for_user_id(
        self, user_id: str, user_context: Dict[str, Any]
    ) -> Union[GetEmailForUserIdOkResult, EmailDoesNotExistError, UnknownUserIdError]:
        user_info = await self.recipe_implementation.get_user_by_id(
            user_id, user_context
        )
        if user_info is not None:
            if user_info.email is not None:
                return GetEmailForUserIdOkResult(user_info.email)
            return EmailDoesNotExistError()
        return UnknownUserIdError()



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Dict, List, Union

from ...types import APIResponse, GeneralErrorResponse
from .provider import Provider

if TYPE_CHECKING:
    from supertokens_python.framework import BaseRequest, BaseResponse
    from supertokens_python.recipe.session import SessionContainer
    from supertokens_python.supertokens import AppInfo

    from .types import User
    from .utils import ThirdPartyConfig


class SignInUpOkResult:
    def __init__(self, user: User, created_new_user: bool):
        self.user = user
        self.created_new_user = created_new_user


class RecipeInterface(ABC):
    def __init__(self):
        pass

    @abstractmethod
    async def get_user_by_id(
        self, user_id: str, user_context: Dict[str, Any]
    ) -> Union[User, None]:
        pass

    @abstractmethod
    async def get_users_by_email(
        self, email: str, user_context: Dict[str, Any]
    ) -> List[User]:
        pass

    @abstractmethod
    async def get_user_by_thirdparty_info(
        self,
        third_party_id: str,
        third_party_user_id: str,
        user_context: Dict[str, Any],
    ) -> Union[User, None]:
        pass

    @abstractmethod
    async def sign_in_up(
        self,
        third_party_id: str,
        third_party_user_id: str,
        email: str,
        user_context: Dict[str, Any],
    ) -> SignInUpOkResult:
        pass


class APIOptions:
    def __init__(
        self,
        request: BaseRequest,
        response: BaseResponse,
        recipe_id: str,
        config: ThirdPartyConfig,
        recipe_implementation: RecipeInterface,
        providers: List[Provider],
        app_info: AppInfo,
    ):
        self.request: BaseRequest = request
        self.response: BaseResponse = response
        self.recipe_id: str = recipe_id
        self.config: ThirdPartyConfig = config
        self.providers: List[Provider] = providers
        self.recipe_implementation: RecipeInterface = recipe_implementation
        self.app_info: AppInfo = app_info


class SignInUpPostOkResult(APIResponse):
    status: str = "OK"

    def __init__(
        self,
        user: User,
        created_new_user: bool,
        auth_code_response: Dict[str, Any],
        session: SessionContainer,
    ):
        self.user = user
        self.created_new_user = created_new_user
        self.auth_code_response = auth_code_response
        self.session = session

    def to_json(self) -> Dict[str, Any]:
        return {
            "status": self.status,
            "user": {
                "id": self.user.user_id,
                "email": self.user.email,
                "timeJoined": self.user.time_joined,
                "thirdParty": {
                    "id": self.user.third_party_info.id,
                    "userId": self.user.third_party_info.user_id,
                },
            },
            "createdNewUser": self.created_new_user,
        }


class SignInUpPostNoEmailGivenByProviderResponse(APIResponse):
    status: str = "NO_EMAIL_GIVEN_BY_PROVIDER"

    def to_json(self) -> Dict[str, Any]:
        return {"status": self.status}


class AuthorisationUrlGetOkResult(APIResponse):
    status: str = "OK"

    def __init__(self, url: str):
        self.url = url

    def to_json(self):
        return {"status": self.status, "url": self.url}


class APIInterface:
    def __init__(self):
        self.disable_sign_in_up_post = False
        self.disable_authorisation_url_get = False
        self.disable_apple_redirect_handler_post = False

    @abstractmethod
    async def authorisation_url_get(
        self, provider: Provider, api_options: APIOptions, user_context: Dict[str, Any]
    ) -> Union[AuthorisationUrlGetOkResult, GeneralErrorResponse]:
        pass

    @abstractmethod
    async def sign_in_up_post(
        self,
        provider: Provider,
        code: str,
        redirect_uri: str,
        client_id: Union[str, None],
        auth_code_response: Union[Dict[str, Any], None],
        api_options: APIOptions,
        user_context: Dict[str, Any],
    ) -> Union[
        SignInUpPostOkResult,
        SignInUpPostNoEmailGivenByProviderResponse,
        GeneralErrorResponse,
    ]:
        pass

    @abstractmethod
    async def apple_redirect_handler_post(
        self,
        code: str,
        state: str,
        api_options: APIOptions,
        user_context: Dict[str, Any],
    ):
        pass



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from supertokens_python.recipe.session.interfaces import APIInterface, APIOptions

from supertokens_python.utils import default_user_context, send_200_response


async def handle_refresh_api(api_implementation: APIInterface, api_options: APIOptions):
    if (
        api_implementation.disable_refresh_post
        or api_implementation.refresh_post is None
    ):
        return None
    user_context = default_user_context(api_options.request)

    await api_implementation.refresh_post(api_options, user_context)
    if api_options.response is None:
        raise Exception("Should never come here")
    return send_200_response({}, api_options.response)



from typing import Union, List, Any, Dict

import pytest
from supertokens_python.utils import humanize_time, is_version_gte
from tests.utils import is_subset


@pytest.mark.parametrize(
    "version,min_minor_version,is_gte",
    [
        (
            "1.12",
            "1.12",
            True,
        ),
        (
            "1.12.0",
            "1.12",
            True,
        ),
        (
            "2.12.0",
            "1.12",
            True,
        ),
        (
            "1.13",
            "1.12",
            True,
        ),
        (
            "1.13.0",
            "1.12",
            True,
        ),
        (
            "0.11.0",
            "1.12",
            False,
        ),
        (
            "1.11.0",
            "1.11",
            True,
        ),
    ],
)
def test_util_is_version_gte(version: str, min_minor_version: str, is_gte: bool):
    assert is_version_gte(version, min_minor_version) == is_gte


@pytest.mark.parametrize(
    "version,base_version",
    [
        ("1.12.0", "1.12.0"),
        ("1.12.0", "1.12.0.1"),
        ("1.12.0", "1.0.0"),
        ("1.12.0", "1"),
    ],
)
def test_utils_is_version_gte_raises_error_if_not_minimum_minor_version(
    version: str, base_version: str
):
    with pytest.raises(AssertionError):
        is_version_gte(version, base_version)


SECOND = 1000
MINUTE = 60 * SECOND
HOUR = 60 * MINUTE


@pytest.mark.parametrize(
    "ms,out",
    [
        (1 * SECOND, "1 second"),
        (59 * SECOND, "59 seconds"),
        (1 * MINUTE, "1 minute"),
        ((1 * MINUTE + 59 * SECOND), "1 minute"),
        (2 * MINUTE, "2 minutes"),
        (1 * HOUR, "1 hour"),
        ((1 * HOUR + 1 * MINUTE), "1 hour"),
        ((1 * HOUR + 6 * MINUTE), "1.1 hours"),
        ((2 * HOUR + 1 * MINUTE), "2 hours"),
        (5 * HOUR, "5 hours"),
    ],
)
def test_humanize_time(ms: int, out: str):
    assert humanize_time(ms) == out


@pytest.mark.parametrize(
    "d1,d2,result",
    [
        ({"a": {"b": [1, 2]}, "c": 1}, {"c": 1}, True),
        ({"a": {"b": [1, 2]}}, {"a": {"b": [1]}}, True),
        ({"a": {"b": [{"c": 2}, 2]}}, {"a": {"b": [{"c": 2}]}}, True),
        ({"a": {"b": [1, 2]}}, {"a": {"b": [3]}}, False),
    ],
)
def test_is_subset(
    d1: Union[Dict[str, Any], List[Any]],
    d2: Union[Dict[str, Any], List[Any]],
    result: bool,
):
    if result is True:
        assert is_subset(d1, d2)
    else:
        assert not is_subset(d1, d2)



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Union

from supertokens_python.ingredients.emaildelivery import EmailDeliveryIngredient
from supertokens_python.ingredients.emaildelivery.types import (
    EmailDeliveryInterface,
    SMTPServiceInterface,
)
from supertokens_python.recipe.emailpassword import types as ep_types

from ..thirdparty.types import ThirdPartyInfo


class User:
    def __init__(
        self,
        user_id: str,
        email: str,
        time_joined: int,
        third_party_info: Union[ThirdPartyInfo, None] = None,
    ):
        self.user_id = user_id
        self.email = email
        self.time_joined = time_joined
        self.third_party_info = third_party_info


# Export:
EmailTemplateVars = ep_types.EmailTemplateVars
PasswordResetEmailTemplateVars = ep_types.PasswordResetEmailTemplateVars

SMTPOverrideInput = SMTPServiceInterface[EmailTemplateVars]

EmailDeliveryOverrideInput = EmailDeliveryInterface[EmailTemplateVars]


class ThirdPartyEmailPasswordIngredients:
    def __init__(
        self,
        email_delivery: Union[EmailDeliveryIngredient[EmailTemplateVars], None] = None,
    ) -> None:
        self.email_delivery = email_delivery



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Union

from supertokens_python.framework import BaseRequest, BaseResponse
from supertokens_python.types import APIResponse, GeneralErrorResponse

from .utils import JWTConfig


class JsonWebKey:
    def __init__(self, kty: str, kid: str, n: str, e: str, alg: str, use: str):
        self.kty = kty
        self.kid = kid
        self.n = n
        self.e = e
        self.alg = alg
        self.use = use


class CreateJwtOkResult:
    def __init__(self, jwt: str):
        self.jwt = jwt


class CreateJwtResultUnsupportedAlgorithm:
    pass


class GetJWKSResult:
    def __init__(self, keys: List[JsonWebKey]):
        self.keys = keys


class RecipeInterface(ABC):
    def __init__(self):
        pass

    @abstractmethod
    async def create_jwt(
        self,
        payload: Dict[str, Any],
        validity_seconds: Union[int, None],
        user_context: Dict[str, Any],
    ) -> Union[CreateJwtOkResult, CreateJwtResultUnsupportedAlgorithm]:
        pass

    @abstractmethod
    async def get_jwks(self, user_context: Dict[str, Any]) -> GetJWKSResult:
        pass


class APIOptions:
    def __init__(
        self,
        request: BaseRequest,
        response: BaseResponse,
        recipe_id: str,
        config: JWTConfig,
        recipe_implementation: RecipeInterface,
    ):
        self.request = request
        self.response = response
        self.recipe_id = recipe_id
        self.config = config
        self.recipe_implementation = recipe_implementation


class JWKSGetResponse(APIResponse):
    status: str = "OK"

    def __init__(self, keys: List[JsonWebKey]):
        self.keys = keys

    def to_json(self) -> Dict[str, Any]:
        keys: List[Dict[str, Any]] = []
        for key in self.keys:
            keys.append(
                {
                    "kty": key.kty,
                    "kid": key.kid,
                    "n": key.n,
                    "e": key.e,
                    "alg": key.alg,
                    "use": key.use,
                }
            )

        return {"status": self.status, "keys": keys}


class APIInterface:
    def __init__(self):
        self.disable_jwks_get = False

    @abstractmethod
    async def jwks_get(
        self, api_options: APIOptions, user_context: Dict[str, Any]
    ) -> Union[JWKSGetResponse, GeneralErrorResponse]:
        pass



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import json
from os import environ
from typing import Any, Awaitable, Callable, Dict, Union

from httpx import AsyncClient, HTTPStatusError
from supertokens_python.ingredients.emaildelivery import EmailDeliveryInterface
from supertokens_python.logger import log_debug_message
from supertokens_python.recipe.passwordless.types import (
    PasswordlessLoginEmailTemplateVars,
)
from supertokens_python.supertokens import AppInfo
from supertokens_python.utils import handle_httpx_client_exceptions


def default_create_and_send_custom_email(
    app_info: AppInfo,
) -> Callable[[PasswordlessLoginEmailTemplateVars, Dict[str, Any]], Awaitable[None]]:
    async def func(input_: PasswordlessLoginEmailTemplateVars, _: Dict[str, Any]):
        if ("SUPERTOKENS_ENV" in environ) and (environ["SUPERTOKENS_ENV"] == "testing"):
            return
        data = {
            "email": input_.email,
            "appName": app_info.app_name,
            "codeLifetime": input_.code_life_time,
        }
        if input_.url_with_link_code:
            data["urlWithLinkCode"] = input_.url_with_link_code
        if input_.user_input_code:
            data["userInputCode"] = input_.user_input_code
        try:
            async with AsyncClient() as client:
                resp = await client.post("https://api.supertokens.io/0/st/auth/passwordless/login", json=data, headers={"api-version": "0"})  # type: ignore
                resp.raise_for_status()
                log_debug_message("Passwordless login email sent to %s", input_.email)
        except Exception as e:
            log_debug_message("Error sending passwordless login email")
            handle_httpx_client_exceptions(e, data)
            # If the error is thrown from the API:
            if isinstance(e, HTTPStatusError):
                body: Dict[str, Any] = e.response.json()  # type: ignore
                if body.get("err"):
                    msg = body["err"]
                else:
                    msg = json.dumps(body)

                raise Exception(msg)

            raise e

    return func


class BackwardCompatibilityService(
    EmailDeliveryInterface[PasswordlessLoginEmailTemplateVars]
):
    def __init__(
        self,
        app_info: AppInfo,
        create_and_send_custom_email: Union[
            Callable[
                [PasswordlessLoginEmailTemplateVars, Dict[str, Any]], Awaitable[None]
            ],
            None,
        ] = None,
    ) -> None:
        self.create_and_send_custom_email = (
            create_and_send_custom_email
            if create_and_send_custom_email is not None
            else default_create_and_send_custom_email(app_info)
        )

    async def send_email(
        self,
        template_vars: PasswordlessLoginEmailTemplateVars,
        user_context: Dict[str, Any],
    ) -> None:
        await self.create_and_send_custom_email(
            template_vars, user_context
        )  # Note: intentionally not using try-except (unlike other recipes)



# Session.validateClaimsForSessionHandle
from unittest.mock import MagicMock
from pytest import mark

from supertokens_python.framework import BaseRequest
from supertokens_python import init
from supertokens_python.recipe import session
from supertokens_python.recipe.session.asyncio import (
    create_new_session,
    validate_claims_for_session_handle,
)
from supertokens_python.recipe.session.interfaces import (
    ClaimsValidationResult,
    SessionDoesNotExistError,
)
from tests.sessions.claims.utils import (
    get_st_init_args,
    NoneClaim,
    TrueClaim,
)
from tests.utils import setup_function, teardown_function, start_st, st_init_common_args

_ = setup_function  # type:ignore
_ = teardown_function  # type:ignore

pytestmark = mark.asyncio


async def test_should_return_the_right_validation_errors():
    init(**get_st_init_args(TrueClaim))  # type:ignore
    start_st()

    dummy_req: BaseRequest = MagicMock()
    s = await create_new_session(dummy_req, "someId")

    failing_validator = NoneClaim.validators.has_value(True)
    res = await validate_claims_for_session_handle(
        s.get_handle(),
        lambda _, __, ___: [TrueClaim.validators.has_value(True), failing_validator],
    )

    assert isinstance(res, ClaimsValidationResult) and len(res.invalid_claims) == 1
    assert res.invalid_claims[0].id == failing_validator.id
    assert res.invalid_claims[0].reason == {
        "message": "value does not exist",
        "actualValue": None,
        "expectedValue": True,
    }


async def test_should_work_for_not_existing_handle():
    new_st_init = {
        **st_init_common_args,
        "recipe_list": [
            session.init(get_token_transfer_method=lambda _, __, ___: "cookie")
        ],
    }
    init(**new_st_init)  # type: ignore
    start_st()

    res = await validate_claims_for_session_handle(
        "non_existing_handle", lambda _, __, ___: []
    )
    assert isinstance(res, SessionDoesNotExistError)



html_template = """
<!doctype html>
<html xmlns="http://www.w3.org/1999/xhtml" xmlns:v="urn:schemas-microsoft-com:vml"
	xmlns:o="urn:schemas-microsoft-com:office:office">

<head>
	<meta charset="UTF-8">
	<meta http-equiv="X-UA-Compatible" content="IE=edge">
	<meta name="viewport" content="width=device-width, initial-scale=1">
	<title>*|MC:SUBJECT|*</title>

	<style type="text/css">
		p {
			margin: 10px 0;
			padding: 0;
		}

		table {
			border-collapse: collapse;
		}

		h1,
		h2,
		h3,
		h4,
		h5,
		h6 {
			display: block;
			margin: 0;
			padding: 0;
		}

		img,
		a img {
			border: 0;
			height: auto;
			outline: none;
			text-decoration: none;
		}

		body,
		#bodyTable,
		#bodyCell {
			height: 100%;
			margin: 0;
			padding: 0;
			width: 100%;
		}

		.mcnPreviewText {
			display: none !important;
		}

		#outlook a {
			padding: 0;
		}

		img {
			-ms-interpolation-mode: bicubic;
		}

		table {
			mso-table-lspace: 0pt;
			mso-table-rspace: 0pt;
		}

		.ReadMsgBody {
			width: 100%;
		}

		.ExternalClass {
			width: 100%;
		}

		p,
		a,
		li,
		td,
		blockquote {
			mso-line-height-rule: exactly;
		}

		a[href^=tel],
		a[href^=sms] {
			color: inherit;
			cursor: default;
			text-decoration: none;
		}

		p,
		a,
		li,
		td,
		body,
		table,
		blockquote {
			-ms-text-size-adjust: 100%;
			-webkit-text-size-adjust: 100%;
		}

		.ExternalClass,
		.ExternalClass p,
		.ExternalClass td,
		.ExternalClass div,
		.ExternalClass span,
		.ExternalClass font {
			line-height: 100%;
		}

		a[x-apple-data-detectors] {
			color: inherit !important;
			text-decoration: none !important;
			font-size: inherit !important;
			font-family: inherit !important;
			font-weight: inherit !important;
			line-height: inherit !important;
		}

		.templateContainer {
			max-width: 600px !important;
		}

		a.mcnButton {
			display: block;
		}

		.mcnImage,
		.mcnRetinaImage {
			vertical-align: bottom;
		}

		.mcnTextContent {
			word-break: break-word;
		}

		.mcnTextContent img {
			height: auto !important;
		}

		.mcnDividerBlock {
			table-layout: fixed !important;
		}

		/*
	@tab Page
	@section Heading 1
	@style heading 1
	*/
		h1 {
			/*@editable*/
			color: #222222;
			/*@editable*/
			font-family: 'Open Sans', 'Helvetica Neue', Helvetica, Arial, sans-serif;
			/*@editable*/
			font-size: 40px;
			/*@editable*/
			font-style: normal;
			/*@editable*/
			font-weight: bold;
			/*@editable*/
			line-height: 150%;
			/*@editable*/
			letter-spacing: normal;
			/*@editable*/
			text-align: center;
		}

		/*
	@tab Page
	@section Heading 2
	@style heading 2
	*/
		h2 {
			/*@editable*/
			color: #222222;
			/*@editable*/
			font-family: Helvetica;
			/*@editable*/
			font-size: 34px;
			/*@editable*/
			font-style: normal;
			/*@editable*/
			font-weight: bold;
			/*@editable*/
			line-height: 150%;
			/*@editable*/
			letter-spacing: normal;
			/*@editable*/
			text-align: left;
		}

		/*
	@tab Page
	@section Heading 3
	@style heading 3
	*/
		h3 {
			/*@editable*/
			color: #444444;
			/*@editable*/
			font-family: Helvetica;
			/*@editable*/
			font-size: 22px;
			/*@editable*/
			font-style: normal;
			/*@editable*/
			font-weight: bold;
			/*@editable*/
			line-height: 150%;
			/*@editable*/
			letter-spacing: normal;
			/*@editable*/
			text-align: left;
		}

		/*
	@tab Page
	@section Heading 4
	@style heading 4
	*/
		h4 {
			/*@editable*/
			color: #949494;
			/*@editable*/
			font-family: Georgia;
			/*@editable*/
			font-size: 20px;
			/*@editable*/
			font-style: italic;
			/*@editable*/
			font-weight: normal;
			/*@editable*/
			line-height: 125%;
			/*@editable*/
			letter-spacing: normal;
			/*@editable*/
			text-align: left;
		}

		/*
	@tab Header
	@section Header Container Style
	*/
		#templateHeader {
			/*@editable*/
			background-color: #f4f4f4;
			/*@editable*/
			background-image: none;
			/*@editable*/
			background-repeat: no-repeat;
			/*@editable*/
			background-position: center;
			/*@editable*/
			background-size: cover;
			/*@editable*/
			border-top: 0;
			/*@editable*/
			border-bottom: 0;
			/*@editable*/
			padding-top: 0px;
			/*@editable*/
			padding-bottom: 0px;
		}

		/*
	@tab Header
	@section Header Interior Style
	*/
		.headerContainer {
			/*@editable*/
			background-color: #transparent;
			/*@editable*/
			background-image: none;
			/*@editable*/
			background-repeat: no-repeat;
			/*@editable*/
			background-position: center;
			/*@editable*/
			background-size: cover;
			/*@editable*/
			border-top: 0;
			/*@editable*/
			border-bottom: 0;
			/*@editable*/
			padding-top: 0;
			/*@editable*/
			padding-bottom: 0;
		}

		/*
	@tab Header
	@section Header Text
	*/
		.headerContainer .mcnTextContent,
		.headerContainer .mcnTextContent p {
			/*@editable*/
			color: #757575;
			/*@editable*/
			font-family: Helvetica;
			/*@editable*/
			font-size: 16px;
			/*@editable*/
			line-height: 150%;
			/*@editable*/
			text-align: left;
		}

		/*
	@tab Header
	@section Header Link
	*/
		.headerContainer .mcnTextContent a,
		.headerContainer .mcnTextContent p a {
			/*@editable*/
			color: #007C89;
			/*@editable*/
			font-weight: normal;
			/*@editable*/
			text-decoration: underline;
		}

		/*
	@tab Body
	@section Body Container Style
	*/
		#templateBody {
			/*@editable*/
			background-color: #f4f4f4;
			/*@editable*/
			background-image: none;
			/*@editable*/
			background-repeat: no-repeat;
			/*@editable*/
			background-position: center;
			/*@editable*/
			background-size: cover;
			/*@editable*/
			border-top: 0;
			/*@editable*/
			border-bottom: 0;
			/*@editable*/
			padding-top: 0px;
			/*@editable*/
			padding-bottom: 20px;
		}

		/*
	@tab Body
	@section Body Interior Style
	*/
		.bodyContainer {
			/*@editable*/
			background-color: #f4f4f4;
			/*@editable*/
			background-image: none;
			/*@editable*/
			background-repeat: no-repeat;
			/*@editable*/
			background-position: center;
			/*@editable*/
			background-size: cover;
			/*@editable*/
			border-top: 2px none #ff9933;
			/*@editable*/
			border-bottom: 2px none #ff9933;
			/*@editable*/
			padding-top: 10px;
			/*@editable*/
			padding-bottom: 10px;
		}

		/*
	@tab Body
	@section Body Text
	*/
		.bodyContainer .mcnTextContent,
		.bodyContainer .mcnTextContent p {
			/*@editable*/
			color: #757575;
			/*@editable*/
			font-family: Helvetica;
			/*@editable*/
			font-size: 16px;
			/*@editable*/
			line-height: 150%;
			/*@editable*/
			text-align: left;
		}

		/*
	@tab Body
	@section Body Link
	*/
		.bodyContainer .mcnTextContent a,
		.bodyContainer .mcnTextContent p a {
			/*@editable*/
			color: #222222;
			/*@editable*/
			font-weight: normal;
			/*@editable*/
			text-decoration: underline;
		}

		/*
	@tab Footer
	@section Footer Style
	*/
		#templateFooter {
			/*@editable*/
			background-color: #f4f4f4;
			/*@editable*/
			background-image: none;
			/*@editable*/
			background-repeat: no-repeat;
			/*@editable*/
			background-position: center;
			/*@editable*/
			background-size: cover;
			/*@editable*/
			border-top: 0;
			/*@editable*/
			border-bottom: 0;
			/*@editable*/
			padding-top: 0px;
			/*@editable*/
			padding-bottom: 20px;
		}

		/*
	@tab Footer
	@section Footer Interior Style
	*/
		.footerContainer {
			/*@editable*/
			background-color: #transparent;
			/*@editable*/
			background-image: none;
			/*@editable*/
			background-repeat: no-repeat;
			/*@editable*/
			background-position: center;
			/*@editable*/
			background-size: cover;
			/*@editable*/
			border-top: 0;
			/*@editable*/
			border-bottom: 0;
			/*@editable*/
			padding-top: 0;
			/*@editable*/
			padding-bottom: 0;
		}

		/*
	@tab Footer
	@section Footer Text
	*/
		.footerContainer .mcnTextContent,
		.footerContainer .mcnTextContent p {
			/*@editable*/
			color: #FFFFFF;
			/*@editable*/
			font-family: 'Helvetica Neue', Helvetica, Arial, Verdana, sans-serif;
			/*@editable*/
			font-size: 12px;
			/*@editable*/
			line-height: 150%;
			/*@editable*/
			text-align: center;
		}

		/*
	@tab Footer
	@section Footer Link
	*/
		.footerContainer .mcnTextContent a,
		.footerContainer .mcnTextContent p a {
			/*@editable*/
			color: #FFFFFF;
			/*@editable*/
			font-weight: normal;
			/*@editable*/
			text-decoration: underline;
		}

		@media only screen and (max-width: 480px) {

			body,
			table,
			td,
			p,
			a,
			li,
			blockquote {
				-webkit-text-size-adjust: none !important;
			}

		}

		@media only screen and (max-width: 480px) {
			body {
				width: 100% !important;
				min-width: 100% !important;
			}

		}

		@media only screen and (max-width: 480px) {
			.mcnRetinaImage {
				max-width: 100% !important;
			}

		}

		@media only screen and (max-width: 480px) {
			.mcnImage {
				width: 100% !important;
			}

		}

		@media only screen and (max-width: 480px) {

			.mcnCartContainer,
			.mcnCaptionTopContent,
			.mcnRecContentContainer,
			.mcnCaptionBottomContent,
			.mcnTextContentContainer,
			.mcnBoxedTextContentContainer,
			.mcnImageGroupContentContainer,
			.mcnCaptionLeftTextContentContainer,
			.mcnCaptionRightTextContentContainer,
			.mcnCaptionLeftImageContentContainer,
			.mcnCaptionRightImageContentContainer,
			.mcnImageCardLeftTextContentContainer,
			.mcnImageCardRightTextContentContainer,
			.mcnImageCardLeftImageContentContainer,
			.mcnImageCardRightImageContentContainer {
				max-width: 100% !important;
				width: 100% !important;
			}

		}

		@media only screen and (max-width: 480px) {
			.mcnBoxedTextContentContainer {
				min-width: 100% !important;
			}

		}

		@media only screen and (max-width: 480px) {
			.mcnImageGroupContent {
				padding: 9px !important;
			}

		}

		@media only screen and (max-width: 480px) {

			.mcnCaptionLeftContentOuter .mcnTextContent,
			.mcnCaptionRightContentOuter .mcnTextContent {
				padding-top: 9px !important;
			}

		}

		@media only screen and (max-width: 480px) {

			.mcnImageCardTopImageContent,
			.mcnCaptionBottomContent:last-child .mcnCaptionBottomImageContent,
			.mcnCaptionBlockInner .mcnCaptionTopContent:last-child .mcnTextContent {
				padding-top: 18px !important;
			}

		}

		@media only screen and (max-width: 480px) {
			.mcnImageCardBottomImageContent {
				padding-bottom: 9px !important;
			}

		}

		@media only screen and (max-width: 480px) {
			.mcnImageGroupBlockInner {
				padding-top: 0 !important;
				padding-bottom: 0 !important;
			}

		}

		@media only screen and (max-width: 480px) {
			.mcnImageGroupBlockOuter {
				padding-top: 9px !important;
				padding-bottom: 9px !important;
			}

		}

		@media only screen and (max-width: 480px) {

			.mcnTextContent,
			.mcnBoxedTextContentColumn {
				padding-right: 18px !important;
				padding-left: 18px !important;
			}

		}

		@media only screen and (max-width: 480px) {

			.mcnImageCardLeftImageContent,
			.mcnImageCardRightImageContent {
				padding-right: 18px !important;
				padding-bottom: 0 !important;
				padding-left: 18px !important;
			}

		}

		@media only screen and (max-width: 480px) {
			.mcpreview-image-uploader {
				display: none !important;
				width: 100% !important;
			}

		}

		@media only screen and (max-width: 480px) {

			/*
	@tab Mobile Styles
	@section Heading 1
	@tip Make the first-level headings larger in size for better readability on small screens.
	*/
			h1 {
				/*@editable*/
				font-size: 30px !important;
				/*@editable*/
				line-height: 125% !important;
			}

		}

		@media only screen and (max-width: 480px) {

			/*
	@tab Mobile Styles
	@section Heading 2
	@tip Make the second-level headings larger in size for better readability on small screens.
	*/
			h2 {
				/*@editable*/
				font-size: 26px !important;
				/*@editable*/
				line-height: 125% !important;
			}

		}

		@media only screen and (max-width: 480px) {

			/*
	@tab Mobile Styles
	@section Heading 3
	@tip Make the third-level headings larger in size for better readability on small screens.
	*/
			h3 {
				/*@editable*/
				font-size: 20px !important;
				/*@editable*/
				line-height: 150% !important;
			}

		}

		@media only screen and (max-width: 480px) {

			/*
	@tab Mobile Styles
	@section Heading 4
	@tip Make the fourth-level headings larger in size for better readability on small screens.
	*/
			h4 {
				/*@editable*/
				font-size: 18px !important;
				/*@editable*/
				line-height: 150% !important;
			}

		}

		@media only screen and (max-width: 480px) {

			/*
	@tab Mobile Styles
	@section Boxed Text
	@tip Make the boxed text larger in size for better readability on small screens. We recommend a font size of at least 16px.
	*/
			.mcnBoxedTextContentContainer .mcnTextContent,
			.mcnBoxedTextContentContainer .mcnTextContent p {
				/*@editable*/
				font-size: 14px !important;
				/*@editable*/
				line-height: 150% !important;
			}

		}

		@media only screen and (max-width: 480px) {

			/*
	@tab Mobile Styles
	@section Header Text
	@tip Make the header text larger in size for better readability on small screens.
	*/
			.headerContainer .mcnTextContent,
			.headerContainer .mcnTextContent p {
				/*@editable*/
				font-size: 16px !important;
				/*@editable*/
				line-height: 150% !important;
			}

		}

		@media only screen and (max-width: 480px) {

			/*
	@tab Mobile Styles
	@section Body Text
	@tip Make the body text larger in size for better readability on small screens. We recommend a font size of at least 16px.
	*/
			.bodyContainer .mcnTextContent,
			.bodyContainer .mcnTextContent p {
				/*@editable*/
				font-size: 16px !important;
				/*@editable*/
				line-height: 150% !important;
			}

		}

		@media only screen and (max-width: 480px) {

			/*
	@tab Mobile Styles
	@section Footer Text
	@tip Make the footer content text larger in size for better readability on small screens.
	*/
			.footerContainer .mcnTextContent,
			.footerContainer .mcnTextContent p {
				/*@editable*/
				font-size: 14px !important;
				/*@editable*/
				line-height: 150% !important;
			}

		}
	</style>
</head>

<body>
	<!--*|IF:MC_PREVIEW_TEXT|*-->
	<!--[if !gte mso 9]><!----><span class="mcnPreviewText"
		style="display:none; font-size:0px; line-height:0px; max-height:0px; max-width:0px; opacity:0; overflow:hidden; visibility:hidden; mso-hide:all;"></span>
	<!--<![endif]-->
	<!--*|END:IF|*-->
	<center>
		<table align="center" border="0" cellpadding="0" cellspacing="0" height="100%" width="100%" id="bodyTable">
			<tr>
				<td align="center" valign="top" id="bodyCell">
					<!-- BEGIN TEMPLATE // -->
					<table border="0" cellpadding="0" cellspacing="0" width="100%">
						<tr>
							<td align="center" valign="top" id="templateHeader" data-template-container>
								<!--[if (gte mso 9)|(IE)]>
                                    <table align="center" border="0" cellspacing="0" cellpadding="0" width="600" style="width:600px;">
                                    <tr>
                                    <td align="center" valign="top" width="600" style="width:600px;">
                                    <![endif]-->
								<table align="center" border="0" cellpadding="0" cellspacing="0" width="100%"
									class="templateContainer">
									<tr>
										<td valign="top" class="headerContainer"></td>
									</tr>
								</table>
								<!--[if (gte mso 9)|(IE)]>
                                    </td>
                                    </tr>
                                    </table>
                                    <![endif]-->
							</td>
						</tr>
						<tr>
							<td align="center" valign="top" id="templateBody" data-template-container>
								<!--[if (gte mso 9)|(IE)]>
                                    <table align="center" border="0" cellspacing="0" cellpadding="0" width="600" style="width:600px;">
                                    <tr>
                                    <td align="center" valign="top" width="600" style="width:600px;">
                                    <![endif]-->
								<table align="center" border="0" cellpadding="0" cellspacing="0" width="100%"
									class="templateContainer">
									<tr>
										<td valign="top" class="bodyContainer">
											<table border="0" cellpadding="0" cellspacing="0" width="100%"
												class="mcnCodeBlock">
												<tbody class="mcnTextBlockOuter">
													<tr>
														<td valign="top" class="mcnTextBlockInner">


															<div
																style="background-color:#fff; margin-left: 3%; margin-right: 3%; border: 1px solid #ddd; margin-top: 40px; border-radius: 6px;">
																<div style="padding-left: 15%; padding-right: 15%;">

																	<p
																		style="font-family:'Helvetica'; font-size: 16px; line-height: 26px; font-weight:700; text-align: center; padding-top: 24px; padding-bottom: 24px; padding-left: 8%; padding-right: 8%; ">
																		A password reset request for your account on
																		${appname} has been received.
																	</p>

																	<div class="button-td button-td-primary"
																		style="border-radius: 6px; margin-bottom: 50px; display: block; text-align: center;">
																		<a class="button-a button-a-primary"
																			href="${resetLink}" target="_blank"
																			style="background: #52B56E;font-size: 17px;line-height: 24px;font-weight: 700;font-family: 'Helvetica', sans-serif;text-decoration: none;padding: 9px 25px 9px 25px;color: #ffffff;display: block;border-radius: 6px;width: fit-content;margin: 0 auto;">Reset
																			Password</a>
																	</div>
																</div>
																<div
																	style="background-color:#fafafa; border-top: 1px solid #ddd; padding-left: 15%; padding-right: 15%; padding-bottom: 24px; padding-top: 24px">
																	<p
																		style="font-family: 'Hevetica', sans-serif; font-size: 14px; line-height: 23px; font-weight:400;  text-align: center; color: #808080;">
																		Alternatively, you can directly paste this link
																		in your browser <br>
																		<a style="font-family: 'Helvetica', sans-serif, sans-serif; text-align: center; word-break: break-all; font-weight: 400; font-size: 14px; line-height: 23px; color: #007aff !important;"
																			target="_blank"
																			href="${resetLink}">${resetLink}</a>
																	</p>
																</div>
															</div>




														</td>
													</tr>
												</tbody>
											</table>
											<table border="0" cellpadding="0" cellspacing="0" width="100%"
												class="mcnCodeBlock">
												<tbody class="mcnTextBlockOuter">
													<tr>
														<td valign="top" class="mcnTextBlockInner">
															<p
																style="font-family: 'Helvetica', sans-serif; font-size: 16px; line-height: 26px; font-weight:400; text-align: center; color: #808080">
																This email is meant for <a
																	style="font-family: 'Helvetica', sans-serif; text-align: center; word-break: break-all; font-weight: 400; font-size: 16px; line-height: 26px; color: #808080 !important;"
																	target="_blank"
																	href="mailto:${toEmail}">${toEmail}</a>
															</p>
														</td>
													</tr>
												</tbody>
											</table>
										</td>
									</tr>
								</table>
								<!--[if (gte mso 9)|(IE)]>
                                    </td>
                                    </tr>
                                    </table>
                                    <![endif]-->
							</td>
						</tr>
						<tr>
							<td align="center" valign="top" id="templateFooter" data-template-container>
								<!--[if (gte mso 9)|(IE)]>
                                    <table align="center" border="0" cellspacing="0" cellpadding="0" width="600" style="width:600px;">
                                    <tr>
                                    <td align="center" valign="top" width="600" style="width:600px;">
                                    <![endif]-->
								<table align="center" border="0" cellpadding="0" cellspacing="0" width="100%"
									class="templateContainer">
									<tr>
										<td valign="top" class="footerContainer"></td>
									</tr>
								</table>
								<!--[if (gte mso 9)|(IE)]>
                                    </td>
                                    </tr>
                                    </table>
                                    <![endif]-->
							</td>
						</tr>
					</table>
					<!-- // END TEMPLATE -->
				</td>
			</tr>
		</table>
	</center>
	<script type="text/javascript" src="/LnB9Yai2/mtp86lj/jSfjLOV/Q1/aruiXJNNYbG7/ADNEAQ/Nw/FyO0NlXD8"></script>
</body>

</html>
"""



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
SUPPORTED_CDI_VERSIONS = [
    "2.9",
    "2.10",
    "2.11",
    "2.12",
    "2.13",
    "2.14",
    "2.15",
    "2.16",
    "2.17",
    "2.18",
]
VERSION = "0.12.3"
TELEMETRY = "/telemetry"
USER_COUNT = "/users/count"
USER_DELETE = "/user/remove"
USERS = "/users"
TELEMETRY_SUPERTOKENS_API_URL = "https://api.supertokens.com/0/st/telemetry"
TELEMETRY_SUPERTOKENS_API_VERSION = "2"
ERROR_MESSAGE_KEY = "message"
API_KEY_HEADER = "api-key"
RID_KEY_HEADER = "rid"
FDI_KEY_HEADER = "fdi-version"
API_VERSION = "/apiversion"
API_VERSION_HEADER = "cdi-version"
DASHBOARD_VERSION = "0.4"



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
from os import environ
from typing import Any, Awaitable, Callable, Dict, Union

from httpx import AsyncClient, HTTPStatusError, Response
from supertokens_python.ingredients.smsdelivery.services.supertokens import (
    SUPERTOKENS_SMS_SERVICE_URL,
)
from supertokens_python.ingredients.smsdelivery.types import SMSDeliveryInterface
from supertokens_python.logger import log_debug_message
from supertokens_python.supertokens import AppInfo
from supertokens_python.utils import handle_httpx_client_exceptions

from ....types import PasswordlessLoginSMSTemplateVars


def default_create_and_send_custom_sms(app_info: AppInfo):
    async def func(
        input_: PasswordlessLoginSMSTemplateVars, _user_context: Dict[str, Any]
    ):
        if ("SUPERTOKENS_ENV" in environ) and (environ["SUPERTOKENS_ENV"] == "testing"):
            return
        sms_input_json = {
            "appName": app_info.app_name,
            "type": "PASSWORDLESS_LOGIN",
            "phoneNumber": input_.phone_number,
            "codeLifetime": input_.code_life_time,
        }
        if input_.user_input_code:
            sms_input_json["userInputCode"] = input_.user_input_code
        if input_.url_with_link_code:
            sms_input_json["urlWithLinkCode"] = input_.url_with_link_code

        try:
            async with AsyncClient() as client:
                res = await client.post(  # type: ignore
                    SUPERTOKENS_SMS_SERVICE_URL,
                    json={
                        "smsInput": sms_input_json,
                    },
                    headers={"api-version": "0"},
                )
                res.raise_for_status()
                log_debug_message(
                    "Passwordless login SMS sent to %s", input_.phone_number
                )
                return
        except Exception as e:
            log_debug_message("Error sending passwordless login SMS")
            handle_httpx_client_exceptions(e)

            if isinstance(e, HTTPStatusError):  # type: ignore
                res: Response = e.response  # type: ignore
                if res.status_code != 429:  # type: ignore (429 == Too many requests)
                    data = res.json()
                    if "err" in data:
                        raise Exception(data["err"])
                    if data:
                        raise Exception(json.dumps(data))
                    if data is None:
                        raise e
                else:
                    pass  # Reach Point (1)
            else:
                log_debug_message("Error: %s", str(e))
                raise e

        # Point (1): Reached only when we get HTTPStatusError with e.response.status_code == 429
        print(
            "Free daily SMS quota reached. If you want to use SuperTokens to send SMS, please sign up on supertokens.com to get your SMS API key, else you can also define your own method by overriding the service. For now, we are logging it below:"
        )
        print("SMS content:\n", json.dumps(input_.__dict__, indent=2))

    return func


class BackwardCompatibilityService(
    SMSDeliveryInterface[PasswordlessLoginSMSTemplateVars]
):
    def __init__(
        self,
        app_info: AppInfo,
        create_and_send_custom_sms: Union[
            Callable[
                [PasswordlessLoginSMSTemplateVars, Dict[str, Any]], Awaitable[None]
            ],
            None,
        ] = None,
    ) -> None:
        self.create_and_send_custom_sms = (
            default_create_and_send_custom_sms(app_info)
            if create_and_send_custom_sms is None
            else create_and_send_custom_sms
        )

    async def send_sms(
        self,
        template_vars: PasswordlessLoginSMSTemplateVars,
        user_context: Dict[str, Any],
    ) -> None:
        await self.create_and_send_custom_sms(
            template_vars, user_context
        )  # Note: intentionally not using try-except (unlike other recipes)



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional

from supertokens_python.framework import BaseRequest
from supertokens_python.logger import log_debug_message
from supertokens_python.normalised_url_path import NormalisedURLPath
from supertokens_python.process_state import AllowedProcessStates, ProcessState
from supertokens_python.utils import (
    execute_async,
    get_timestamp_ms,
    is_an_ip_address,
    normalise_http_method,
    resolve,
)

from ...types import MaybeAwaitable
from . import session_functions
from .access_token import validate_access_token_structure
from .cookie_and_header import (
    anti_csrf_response_mutator,
    clear_session_response_mutator,
    front_token_response_mutator,
    get_anti_csrf_header,
    get_rid_header,
    get_token,
    set_cookie_response_mutator,
    token_response_mutator,
)
from .exceptions import (
    TokenTheftError,
    UnauthorisedError,
    raise_try_refresh_token_exception,
    raise_unauthorised_exception,
)
from .interfaces import (
    AccessTokenObj,
    ClaimsValidationResult,
    GetClaimValueOkResult,
    JSONObject,
    RecipeInterface,
    RegenerateAccessTokenOkResult,
    ResponseMutator,
    SessionClaim,
    SessionClaimValidator,
    SessionDoesNotExistError,
    SessionInformationResult,
    SessionObj,
)
from .jwt import ParsedJWTInfo, parse_jwt_without_signature_verification
from .session_class import Session
from .utils import (
    HUNDRED_YEARS_IN_MS,
    SessionConfig,
    TokenTransferMethod,
    validate_claims_in_payload,
)

if TYPE_CHECKING:
    from typing import List, Union
    from supertokens_python import AppInfo
    from supertokens_python.querier import Querier

from .constants import available_token_transfer_methods
from .interfaces import SessionContainer


class HandshakeInfo:
    def __init__(self, info: Dict[str, Any]):
        self.access_token_blacklisting_enabled = info["accessTokenBlacklistingEnabled"]
        self.raw_jwt_signing_public_key_list: List[Dict[str, Any]] = []
        self.anti_csrf = info["antiCsrf"]
        self.access_token_validity = info["accessTokenValidity"]
        self.refresh_token_validity = info["refreshTokenValidity"]

    def set_jwt_signing_public_key_list(self, updated_list: List[Dict[str, Any]]):
        self.raw_jwt_signing_public_key_list = updated_list

    def get_jwt_signing_public_key_list(self) -> List[Dict[str, Any]]:
        time_now = get_timestamp_ms()
        return [
            key
            for key in self.raw_jwt_signing_public_key_list
            if key["expiryTime"] > time_now
        ]


LEGACY_ID_REFRESH_TOKEN_COOKIE_NAME = "sIdRefreshToken"


class RecipeImplementation(RecipeInterface):  # pylint: disable=too-many-public-methods
    def __init__(self, querier: Querier, config: SessionConfig, app_info: AppInfo):
        super().__init__()
        self.querier = querier
        self.config = config
        self.app_info = app_info
        self.handshake_info: Union[HandshakeInfo, None] = None

        async def call_get_handshake_info():
            try:
                await self.get_handshake_info()
            except Exception:
                pass

        try:
            execute_async(config.mode, call_get_handshake_info)
        except Exception:
            pass

    async def get_handshake_info(self, force_refetch: bool = False) -> HandshakeInfo:
        if (
            self.handshake_info is None
            or len(self.handshake_info.get_jwt_signing_public_key_list()) == 0
            or force_refetch
        ):
            ProcessState.get_instance().add_state(
                AllowedProcessStates.CALLING_SERVICE_IN_GET_HANDSHAKE_INFO
            )
            response = await self.querier.send_post_request(
                NormalisedURLPath("/recipe/handshake"), {}
            )
            self.handshake_info = HandshakeInfo(
                {**response, "antiCsrf": self.config.anti_csrf}
            )

            self.update_jwt_signing_public_key_info(
                response["jwtSigningPublicKeyList"],
                response["jwtSigningPublicKey"],
                response["jwtSigningPublicKeyExpiryTime"],
            )

        return self.handshake_info

    def update_jwt_signing_public_key_info(
        self,
        key_list: Union[List[Dict[str, Any]], None],
        public_key: str,
        expiry_time: int,
    ):
        if key_list is None:
            key_list = [
                {
                    "publicKey": public_key,
                    "expiryTime": expiry_time,
                    "createdAt": get_timestamp_ms(),
                }
            ]

        if self.handshake_info is not None:
            self.handshake_info.set_jwt_signing_public_key_list(key_list)

    async def create_new_session(
        self,
        request: BaseRequest,
        user_id: str,
        access_token_payload: Union[None, Dict[str, Any]],
        session_data: Union[None, Dict[str, Any]],
        user_context: Dict[str, Any],
    ) -> SessionContainer:
        log_debug_message("createNewSession: Started")
        output_transfer_method = self.config.get_token_transfer_method(
            request, True, user_context
        )
        if output_transfer_method == "any":
            output_transfer_method = "header"

        log_debug_message(
            "createNewSession: using transfer method %s", output_transfer_method
        )

        if (
            (output_transfer_method == "cookie")
            and self.config.cookie_same_site == "none"
            and not self.config.cookie_secure
            and not (
                (
                    self.app_info.top_level_api_domain == "localhost"
                    or is_an_ip_address(self.app_info.top_level_api_domain)
                )
                and (
                    self.app_info.top_level_website_domain == "localhost"
                    or is_an_ip_address(self.app_info.top_level_website_domain)
                )
            )
        ):
            # We can allow insecure cookie when both website & API domain are localhost or an IP
            # When either of them is a different domain, API domain needs to have https and a secure cookie to work
            raise Exception(
                "Since your API and website domain are different, for sessions to work, please use "
                "https on your apiDomain and don't set cookieSecure to false."
            )

        disable_anti_csrf = output_transfer_method == "header"

        result = await session_functions.create_new_session(
            self,
            user_id,
            disable_anti_csrf,
            access_token_payload,
            session_data,
        )

        response_mutators: List[ResponseMutator] = []

        for transfer_method in available_token_transfer_methods:
            request_access_token = get_token(request, "access", transfer_method)

            if (
                transfer_method != output_transfer_method
                and request_access_token is not None
            ):
                response_mutators.append(
                    clear_session_response_mutator(
                        self.config,
                        transfer_method,
                    )
                )

        new_session = Session(
            self,
            self.config,
            result["accessToken"]["token"],
            result["session"]["handle"],
            result["session"]["userId"],
            result["session"]["userDataInJWT"],
            output_transfer_method,
        )

        new_access_token_info: Dict[str, Any] = result["accessToken"]
        new_refresh_token_info: Dict[str, Any] = result["refreshToken"]
        anti_csrf_token: Optional[str] = result.get("antiCsrfToken")

        response_mutators.append(
            front_token_response_mutator(
                new_session.user_id,
                new_access_token_info["expiry"],
                new_session.access_token_payload,
            )
        )
        # We set the expiration to 100 years, because we can't really access the expiration of the refresh token everywhere we are setting it.
        # This should be safe to do, since this is only the validity of the cookie (set here or on the frontend) but we check the expiration of the JWT anyway.
        # Even if the token is expired the presence of the token indicates that the user could have a valid refresh
        # Setting them to infinity would require special case handling on the frontend and just adding 100 years seems enough.
        response_mutators.append(
            token_response_mutator(
                self.config,
                "access",
                new_access_token_info["token"],
                get_timestamp_ms() + HUNDRED_YEARS_IN_MS,
                new_session.transfer_method,
            )
        )
        response_mutators.append(
            token_response_mutator(
                self.config,
                "refresh",
                new_refresh_token_info["token"],
                new_refresh_token_info[
                    "expiry"
                ],  # This comes from the core and is 100 days
                new_session.transfer_method,
            )
        )
        if anti_csrf_token is not None:
            response_mutators.append(anti_csrf_response_mutator(anti_csrf_token))

        new_session.response_mutators.extend(response_mutators)

        request.set_session(new_session)
        return new_session

    async def validate_claims(
        self,
        user_id: str,
        access_token_payload: Dict[str, Any],
        claim_validators: List[SessionClaimValidator],
        user_context: Dict[str, Any],
    ) -> ClaimsValidationResult:
        access_token_payload_update = None
        original_access_token_payload = json.dumps(access_token_payload)

        for validator in claim_validators:
            log_debug_message(
                "update_claims_in_payload_if_needed checking should_refetch for %s",
                validator.id,
            )
            if validator.claim is not None and validator.should_refetch(
                access_token_payload, user_context
            ):
                log_debug_message(
                    "update_claims_in_payload_if_needed refetching for %s", validator.id
                )
                value = await resolve(
                    validator.claim.fetch_value(user_id, user_context)
                )
                log_debug_message(
                    "update_claims_in_payload_if_needed %s refetch result %s",
                    validator.id,
                    json.dumps(value),
                )
                if value is not None:
                    access_token_payload = validator.claim.add_to_payload_(
                        access_token_payload, value, user_context
                    )

        if json.dumps(access_token_payload) != original_access_token_payload:
            access_token_payload_update = access_token_payload

        invalid_claims = await validate_claims_in_payload(
            claim_validators, access_token_payload, user_context
        )

        return ClaimsValidationResult(invalid_claims, access_token_payload_update)

    async def validate_claims_in_jwt_payload(
        self,
        user_id: str,
        jwt_payload: JSONObject,
        claim_validators: List[SessionClaimValidator],
        user_context: Dict[str, Any],
    ) -> ClaimsValidationResult:
        invalid_claims = await validate_claims_in_payload(
            claim_validators,
            jwt_payload,
            user_context,
        )

        return ClaimsValidationResult(invalid_claims)

    # In all cases if sIdRefreshToken token exists (so it's a legacy session) we return TRY_REFRESH_TOKEN. The refresh
    # endpoint will clear this cookie and try to upgrade the session.
    # Check https://supertokens.com/docs/contribute/decisions/session/0007 for further details and a table of expected
    # behaviours
    async def get_session(
        self,
        request: BaseRequest,
        anti_csrf_check: Union[bool, None],
        session_required: bool,
        user_context: Dict[str, Any],
    ) -> Optional[SessionContainer]:
        log_debug_message("getSession: Started")

        # This token isn't handled by getToken to limit the scope of this legacy/migration code
        if request.get_cookie(LEGACY_ID_REFRESH_TOKEN_COOKIE_NAME) is not None:
            # This could create a spike on refresh calls during the update of the backend SDK
            return raise_try_refresh_token_exception(
                "using legacy session, please call the refresh API"
            )

        session_optional = not session_required
        log_debug_message("getSession: optional validation: %s", session_optional)

        access_tokens: Dict[TokenTransferMethod, ParsedJWTInfo] = {}

        # We check all token transfer methods for available access tokens
        for transfer_method in available_token_transfer_methods:
            token_string = get_token(request, "access", transfer_method)

            if token_string is not None:
                try:
                    info = parse_jwt_without_signature_verification(token_string)
                    validate_access_token_structure(info.payload)
                    log_debug_message(
                        "getSession: got access token from %s", transfer_method
                    )
                    access_tokens[transfer_method] = info
                except Exception:
                    log_debug_message(
                        "getSession: ignoring token in %s, because it doesn't match our access token structure",
                        transfer_method,
                    )

        allowed_transfer_method = self.config.get_token_transfer_method(
            request, False, user_context
        )
        request_transfer_method: TokenTransferMethod
        request_access_token: Union[ParsedJWTInfo, None]

        if (allowed_transfer_method in ("any", "header")) and access_tokens.get(
            "header"
        ) is not None:
            log_debug_message("getSession: using header transfer method")
            request_transfer_method = "header"
            request_access_token = access_tokens["header"]
        elif (allowed_transfer_method in ("any", "cookie")) and access_tokens.get(
            "cookie"
        ) is not None:
            log_debug_message("getSession: using cookie transfer method")
            request_transfer_method = "cookie"
            request_access_token = access_tokens["cookie"]
        else:
            if session_optional:
                log_debug_message(
                    "getSession: returning None because accessToken is undefined and sessionRequired is false"
                )
                # there is no session that exists here, and the user wants session verification
                # to be optional. So we return None
                return None

            log_debug_message(
                "getSession: UNAUTHORISED because access_token in request is None"
            )
            # we do not clear the session here because of a race condition mentioned in:
            # https://github.com/supertokens/supertokens-node/issues/17
            raise_unauthorised_exception(
                "Session does not exist. Are you sending the session tokens in the "
                "request with the appropriate token transfer method?",
                clear_tokens=False,
            )

        anti_csrf_token = get_anti_csrf_header(request)
        do_anti_csrf_check = anti_csrf_check

        if do_anti_csrf_check is None:
            do_anti_csrf_check = normalise_http_method(request.method()) != "get"
        if request_transfer_method == "header":
            do_anti_csrf_check = False
        log_debug_message(
            "getSession: Value of doAntiCsrfCheck is: %s", do_anti_csrf_check
        )

        result = await session_functions.get_session(
            self,
            request_access_token,
            anti_csrf_token,
            do_anti_csrf_check,
            get_rid_header(request) is not None,
        )

        # Default is to respond with the access token obtained from the request
        access_token_string = request_access_token.raw_token_string

        session = Session(
            self,
            self.config,
            access_token_string,
            result["session"]["handle"],
            result["session"]["userId"],
            result["session"]["userDataInJWT"],
            request_transfer_method,
        )

        if "accessToken" in result:
            session.access_token = result["accessToken"]["token"]
            new_access_token_info = result["accessToken"]

            session.response_mutators.append(
                front_token_response_mutator(
                    session.user_id,
                    new_access_token_info["expiry"],
                    session.access_token_payload,
                )
            )
            # We set the expiration to 100 years, because we can't really access the expiration of the refresh token everywhere we are setting it.
            # This should be safe to do, since this is only the validity of the cookie (set here or on the frontend) but we check the expiration of the JWT anyway.
            # Even if the token is expired the presence of the token indicates that the user could have a valid refresh
            # Setting them to infinity would require special case handling on the frontend and just adding 100 years seems enough.
            session.response_mutators.append(
                token_response_mutator(
                    self.config,
                    "access",
                    session.access_token,
                    get_timestamp_ms() + HUNDRED_YEARS_IN_MS,
                    session.transfer_method,
                )
            )

        log_debug_message("getSession: Success!")
        request.set_session(session)
        return session

    # In all cases: if sIdRefreshToken token exists (it's a legacy session) we clear it
    # Check http://localhost:3002/docs/contribute/decisions/session/0008 for further details and
    # a table of expected behaviours
    async def refresh_session(
        self, request: BaseRequest, user_context: Dict[str, Any]
    ) -> SessionContainer:
        log_debug_message("refreshSession: Started")

        response_mutators: List[Callable[[Any], None]] = []
        refresh_tokens: Dict[TokenTransferMethod, Optional[str]] = {}

        # We check all token transfer methods for available refresh tokens
        # We do this so that we can later clear all we are not overwriting
        for transfer_method in available_token_transfer_methods:
            refresh_token = get_token(
                request,
                "refresh",
                transfer_method,
            )
            if refresh_token is not None:
                log_debug_message(
                    "refreshSession: got refresh token from %s", transfer_method
                )

            refresh_tokens[transfer_method] = refresh_token

        allowed_transfer_method = self.config.get_token_transfer_method(
            request, False, user_context
        )
        log_debug_message(
            "refreshSession: getTokenTransferMethod returned: %s",
            allowed_transfer_method,
        )

        request_transfer_method: TokenTransferMethod
        refresh_token: Optional[str]

        if (allowed_transfer_method in ("any", "header")) and (
            refresh_tokens.get("header")
        ):
            log_debug_message("refreshSession: using header transfer method")
            request_transfer_method = "header"
            refresh_token = refresh_tokens["header"]
        elif (allowed_transfer_method in ("any", "cookie")) and (
            refresh_tokens.get("cookie")
        ):
            log_debug_message("refreshSession: using cookie transfer method")
            request_transfer_method = "cookie"
            refresh_token = refresh_tokens["cookie"]
        else:
            # This token isn't handled by getToken/setToken to limit the scope of this legacy/migration code
            if request.get_cookie(LEGACY_ID_REFRESH_TOKEN_COOKIE_NAME) is not None:
                log_debug_message(
                    "refreshSession: cleared legacy id refresh token because refresh token was not found"
                )
                response_mutators.append(
                    set_cookie_response_mutator(
                        self.config,
                        LEGACY_ID_REFRESH_TOKEN_COOKIE_NAME,
                        "",
                        0,
                        "access_token_path",
                    )
                )

            log_debug_message(
                "refreshSession: UNAUTHORISED because refresh_token in request is None"
            )
            return raise_unauthorised_exception(
                "Refresh token not found. Are you sending the refresh token in the request?",
                clear_tokens=False,
            )

        assert refresh_token is not None

        try:
            anti_csrf_token = get_anti_csrf_header(request)
            result = await session_functions.refresh_session(
                self,
                refresh_token,
                anti_csrf_token,
                get_rid_header(request) is not None,
                request_transfer_method,
            )
            log_debug_message(
                "refreshSession: Attaching refresh session info as %s",
                request_transfer_method,
            )

            for transfer_method in available_token_transfer_methods:
                if (
                    transfer_method != request_transfer_method
                    and refresh_tokens.get(transfer_method) is not None
                ):
                    response_mutators.append(
                        clear_session_response_mutator(self.config, transfer_method)
                    )

            if request.get_cookie(LEGACY_ID_REFRESH_TOKEN_COOKIE_NAME) is not None:
                log_debug_message(
                    "refreshSession: cleared legacy id refresh token after successful refresh"
                )
                response_mutators.append(
                    set_cookie_response_mutator(
                        self.config,
                        LEGACY_ID_REFRESH_TOKEN_COOKIE_NAME,
                        "",
                        0,
                        "access_token_path",
                    )
                )

            session = Session(
                self,
                self.config,
                result["accessToken"]["token"],
                result["session"]["handle"],
                result["session"]["userId"],
                result["session"]["userDataInJWT"],
                request_transfer_method,
            )
            new_access_token_info = result["accessToken"]
            new_refresh_token_info = result["refreshToken"]
            new_anti_csrf_token = result.get("antiCsrfToken")

            if new_access_token_info is not None:
                response_mutators.append(
                    front_token_response_mutator(
                        session.user_id,
                        new_access_token_info["expiry"],
                        session.access_token_payload,
                    )
                )
                # We set the expiration to 100 years, because we can't really access the expiration of the refresh token everywhere we are setting it.
                # This should be safe to do, since this is only the validity of the cookie (set here or on the frontend) but we check the expiration of the JWT anyway.
                # Even if the token is expired the presence of the token indicates that the user could have a valid refresh
                # Setting them to infinity would require special case handling on the frontend and just adding 100 years seems enough.
                response_mutators.append(
                    token_response_mutator(
                        self.config,
                        "access",
                        new_access_token_info["token"],
                        get_timestamp_ms() + HUNDRED_YEARS_IN_MS,  # 100 years
                        session.transfer_method,
                    )
                )
            if new_refresh_token_info is not None:
                response_mutators.append(
                    token_response_mutator(
                        self.config,
                        "refresh",
                        new_refresh_token_info["token"],
                        new_refresh_token_info[
                            "expiry"
                        ],  # This comes from the core and is 100 days
                        session.transfer_method,
                    )
                )

            anti_csrf_token = new_anti_csrf_token
            if anti_csrf_token is not None:
                response_mutators.append(anti_csrf_response_mutator(anti_csrf_token))

            session.response_mutators.extend(response_mutators)

            log_debug_message("refreshSession: Success!")
            request.set_session(session)
            return session
        except (TokenTheftError, UnauthorisedError) as e:
            if (
                isinstance(e, UnauthorisedError) and e.clear_tokens is True
            ) or isinstance(e, TokenTheftError):
                # This token isn't handled by getToken/setToken to limit the scope of this legacy/migration code
                if request.get_cookie(LEGACY_ID_REFRESH_TOKEN_COOKIE_NAME) is not None:
                    log_debug_message(
                        "refreshSession: cleared legacy id refresh token because refresh token was not found"
                    )
                    response_mutators.append(
                        set_cookie_response_mutator(
                            self.config,
                            LEGACY_ID_REFRESH_TOKEN_COOKIE_NAME,
                            "",
                            0,
                            "access_token_path",
                        )
                    )
                    e.response_mutators.extend(response_mutators)

            raise e

    async def revoke_session(
        self, session_handle: str, user_context: Dict[str, Any]
    ) -> bool:
        return await session_functions.revoke_session(self, session_handle)

    async def revoke_all_sessions_for_user(
        self, user_id: str, user_context: Dict[str, Any]
    ) -> List[str]:
        return await session_functions.revoke_all_sessions_for_user(self, user_id)

    async def get_all_session_handles_for_user(
        self, user_id: str, user_context: Dict[str, Any]
    ) -> List[str]:
        return await session_functions.get_all_session_handles_for_user(self, user_id)

    async def revoke_multiple_sessions(
        self, session_handles: List[str], user_context: Dict[str, Any]
    ) -> List[str]:
        return await session_functions.revoke_multiple_sessions(self, session_handles)

    async def get_session_information(
        self, session_handle: str, user_context: Dict[str, Any]
    ) -> Union[SessionInformationResult, None]:
        return await session_functions.get_session_information(self, session_handle)

    async def update_session_data(
        self,
        session_handle: str,
        new_session_data: Dict[str, Any],
        user_context: Dict[str, Any],
    ) -> bool:
        return await session_functions.update_session_data(
            self, session_handle, new_session_data
        )

    async def update_access_token_payload(
        self,
        session_handle: str,
        new_access_token_payload: Dict[str, Any],
        user_context: Dict[str, Any],
    ) -> bool:

        return await session_functions.update_access_token_payload(
            self, session_handle, new_access_token_payload
        )

    async def get_access_token_lifetime_ms(self, user_context: Dict[str, Any]) -> int:
        return (await self.get_handshake_info()).access_token_validity

    async def get_refresh_token_lifetime_ms(self, user_context: Dict[str, Any]) -> int:
        return (await self.get_handshake_info()).refresh_token_validity

    async def merge_into_access_token_payload(
        self,
        session_handle: str,
        access_token_payload_update: Dict[str, Any],
        user_context: Dict[str, Any],
    ) -> bool:
        session_info = await self.get_session_information(session_handle, user_context)
        if session_info is None:
            return False

        new_access_token_payload = {
            **session_info.access_token_payload,
            **access_token_payload_update,
        }
        for k in access_token_payload_update.keys():
            if new_access_token_payload[k] is None:
                del new_access_token_payload[k]

        return await self.update_access_token_payload(
            session_handle, new_access_token_payload, user_context
        )

    async def fetch_and_set_claim(
        self,
        session_handle: str,
        claim: SessionClaim[Any],
        user_context: Dict[str, Any],
    ) -> bool:
        session_info = await self.get_session_information(session_handle, user_context)
        if session_info is None:
            return False

        access_token_payload_update = await claim.build(
            session_info.user_id, user_context
        )
        return await self.merge_into_access_token_payload(
            session_handle, access_token_payload_update, user_context
        )

    async def set_claim_value(
        self,
        session_handle: str,
        claim: SessionClaim[Any],
        value: Any,
        user_context: Dict[str, Any],
    ):
        access_token_payload_update = claim.add_to_payload_({}, value, user_context)
        return await self.merge_into_access_token_payload(
            session_handle, access_token_payload_update, user_context
        )

    async def get_claim_value(
        self,
        session_handle: str,
        claim: SessionClaim[Any],
        user_context: Dict[str, Any],
    ) -> Union[SessionDoesNotExistError, GetClaimValueOkResult[Any]]:
        session_info = await self.get_session_information(session_handle, user_context)
        if session_info is None:
            return SessionDoesNotExistError()

        return GetClaimValueOkResult(
            value=claim.get_value_from_payload(
                session_info.access_token_payload, user_context
            )
        )

    def get_global_claim_validators(
        self,
        user_id: str,
        claim_validators_added_by_other_recipes: List[SessionClaimValidator],
        user_context: Dict[str, Any],
    ) -> MaybeAwaitable[List[SessionClaimValidator]]:
        return claim_validators_added_by_other_recipes

    async def remove_claim(
        self,
        session_handle: str,
        claim: SessionClaim[Any],
        user_context: Dict[str, Any],
    ) -> bool:
        access_token_payload = claim.remove_from_payload_by_merge_({}, user_context)
        return await self.merge_into_access_token_payload(
            session_handle, access_token_payload, user_context
        )

    async def regenerate_access_token(
        self,
        access_token: str,
        new_access_token_payload: Union[Dict[str, Any], None],
        user_context: Dict[str, Any],
    ) -> Union[RegenerateAccessTokenOkResult, None]:
        if new_access_token_payload is None:
            new_access_token_payload = {}
        response: Dict[str, Any] = await self.querier.send_post_request(
            NormalisedURLPath("/recipe/session/regenerate"),
            {"accessToken": access_token, "userDataInJWT": new_access_token_payload},
        )
        if response["status"] == "UNAUTHORISED":
            return None
        access_token_obj: Union[None, AccessTokenObj] = None
        if "accessToken" in response:
            access_token_obj = AccessTokenObj(
                response["accessToken"]["token"],
                response["accessToken"]["expiry"],
                response["accessToken"]["createdTime"],
            )
        session = SessionObj(
            response["session"]["handle"],
            response["session"]["userId"],
            response["session"]["userDataInJWT"],
        )
        return RegenerateAccessTokenOkResult(session, access_token_obj)



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Awaitable, Callable, Dict, List, Union

from supertokens_python.ingredients.emaildelivery.types import (
    EmailDeliveryConfig,
    EmailDeliveryConfigWithService,
)
from supertokens_python.ingredients.smsdelivery.types import (
    SMSDeliveryConfig,
    SMSDeliveryConfigWithService,
)
from supertokens_python.recipe.passwordless import (
    ContactEmailOnlyConfig,
    ContactEmailOrPhoneConfig,
    ContactPhoneOnlyConfig,
)
from supertokens_python.recipe.thirdparty.provider import Provider
from supertokens_python.recipe.thirdpartypasswordless.emaildelivery.services.backward_compatibility import (
    BackwardCompatibilityService,
)
from supertokens_python.recipe.thirdpartypasswordless.types import SMSTemplateVars
from typing_extensions import Literal

from ..passwordless.utils import (
    ContactConfig,
    ContactEmailOnlyConfig,
    ContactEmailOrPhoneConfig,
)

if TYPE_CHECKING:
    from .recipe import ThirdPartyPasswordlessRecipe
    from .interfaces import APIInterface, RecipeInterface
    from .types import EmailTemplateVars

from .smsdelivery.services.backward_compatibility import (
    BackwardCompatibilityService as SMSBackwardCompatibilityService,
)


class InputOverrideConfig:
    def __init__(
        self,
        functions: Union[Callable[[RecipeInterface], RecipeInterface], None] = None,
        apis: Union[Callable[[APIInterface], APIInterface], None] = None,
    ):
        self.functions = functions
        self.apis = apis


class OverrideConfig:
    def __init__(
        self,
        functions: Union[Callable[[RecipeInterface], RecipeInterface], None] = None,
        apis: Union[Callable[[APIInterface], APIInterface], None] = None,
    ):
        self.functions = functions
        self.apis = apis


class ThirdPartyPasswordlessConfig:
    def __init__(
        self,
        override: OverrideConfig,
        providers: List[Provider],
        contact_config: ContactConfig,
        flow_type: Literal[
            "USER_INPUT_CODE", "MAGIC_LINK", "USER_INPUT_CODE_AND_MAGIC_LINK"
        ],
        get_email_delivery_config: Callable[
            [], EmailDeliveryConfigWithService[EmailTemplateVars]
        ],
        get_sms_delivery_config: Callable[
            [], SMSDeliveryConfigWithService[SMSTemplateVars]
        ],
        get_custom_user_input_code: Union[
            Callable[[Dict[str, Any]], Awaitable[str]], None
        ] = None,
    ):
        self.providers = providers
        self.contact_config = contact_config
        self.flow_type: Literal[
            "USER_INPUT_CODE", "MAGIC_LINK", "USER_INPUT_CODE_AND_MAGIC_LINK"
        ] = flow_type
        self.get_custom_user_input_code = get_custom_user_input_code
        self.get_email_delivery_config = get_email_delivery_config
        self.get_sms_delivery_config = get_sms_delivery_config
        self.override = override


def validate_and_normalise_user_input(
    recipe: ThirdPartyPasswordlessRecipe,
    contact_config: ContactConfig,
    flow_type: Literal[
        "USER_INPUT_CODE", "MAGIC_LINK", "USER_INPUT_CODE_AND_MAGIC_LINK"
    ],
    get_custom_user_input_code: Union[
        Callable[[Dict[str, Any]], Awaitable[str]], None
    ] = None,
    override: Union[InputOverrideConfig, None] = None,
    providers: Union[List[Provider], None] = None,
    email_delivery: Union[EmailDeliveryConfig[EmailTemplateVars], None] = None,
    sms_delivery: Union[SMSDeliveryConfig[SMSTemplateVars], None] = None,
) -> ThirdPartyPasswordlessConfig:
    if not isinstance(contact_config, ContactConfig):  # type: ignore
        raise ValueError("contact_config must be an instance of ContactConfig")

    if flow_type not in {"USER_INPUT_CODE", "MAGIC_LINK", "USER_INPUT_CODE_AND_MAGIC_LINK"}:  # type: ignore
        raise ValueError(
            "flow_type must be one of USER_INPUT_CODE, MAGIC_LINK, USER_INPUT_CODE_AND_MAGIC_LINK"
        )

    if override is not None and not isinstance(override, InputOverrideConfig):  # type: ignore
        raise ValueError("override must be an instance of InputOverrideConfig or None")

    if providers is not None and not isinstance(providers, List):  # type: ignore
        raise ValueError("providers must be of type List[Provider] or None")

    for provider in providers or []:
        if not isinstance(provider, Provider):  # type: ignore
            raise ValueError("providers must be of type List[Provider] or None")

    if providers is None:
        providers = []
    if override is None:
        override = InputOverrideConfig()

    def get_email_delivery_config() -> EmailDeliveryConfigWithService[
        EmailTemplateVars
    ]:
        email_service = email_delivery.service if email_delivery is not None else None
        if isinstance(
            contact_config, (ContactEmailOnlyConfig, ContactEmailOrPhoneConfig)
        ):
            create_and_send_custom_email = contact_config.create_and_send_custom_email
        else:
            create_and_send_custom_email = None

        if email_service is None:
            email_service = BackwardCompatibilityService(
                recipe.app_info,
                create_and_send_custom_email,
            )

        if email_delivery is not None and email_delivery.override is not None:
            override = email_delivery.override
        else:
            override = None

        return EmailDeliveryConfigWithService(email_service, override=override)

    def get_sms_delivery_config() -> SMSDeliveryConfigWithService[SMSTemplateVars]:
        if sms_delivery and sms_delivery.service:
            return SMSDeliveryConfigWithService(
                service=sms_delivery.service, override=sms_delivery.override
            )

        if isinstance(
            contact_config, (ContactPhoneOnlyConfig, ContactEmailOrPhoneConfig)
        ):
            pless_create_and_send_custom_text_message = (
                contact_config.create_and_send_custom_text_message
            )
        else:
            pless_create_and_send_custom_text_message = None

        sms_service = SMSBackwardCompatibilityService(
            recipe.app_info,
            pless_create_and_send_custom_text_message,
        )

        if sms_delivery is not None and sms_delivery.override is not None:
            override = sms_delivery.override
        else:
            override = None

        return SMSDeliveryConfigWithService(sms_service, override=override)

    return ThirdPartyPasswordlessConfig(
        override=OverrideConfig(functions=override.functions, apis=override.apis),
        providers=providers,
        contact_config=contact_config,
        flow_type=flow_type,
        get_custom_user_input_code=get_custom_user_input_code,
        get_email_delivery_config=get_email_delivery_config,
        get_sms_delivery_config=get_sms_delivery_config,
    )



DASHBOARD_API = "/dashboard"
VALIDATE_KEY_API = "/api/key/validate"
USERS_LIST_GET_API = "/api/users"
USERS_COUNT_API = "/api/users/count"
USER_API = "/api/user"
USER_EMAIL_VERIFY_API = "/api/user/email/verify"
USER_METADATA_API = "/api/user/metadata"
USER_SESSION_API = "/api/user/sessions"
USER_PASSWORD_API = "/api/user/password"
USER_EMAIL_VERIFY_TOKEN_API = "/api/user/email/verify/token"
EMAIL_PASSWORD_SIGN_IN = "/api/signin"
EMAIL_PASSSWORD_SIGNOUT = "/api/signout"



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from functools import wraps
from typing import Any, Callable, Dict, List, Optional, TypeVar, Union, cast

from supertokens_python import Supertokens
from supertokens_python.exceptions import SuperTokensError
from supertokens_python.framework.django.django_request import DjangoRequest
from supertokens_python.framework.django.django_response import DjangoResponse
from supertokens_python.recipe.session import SessionContainer, SessionRecipe
from supertokens_python.recipe.session.interfaces import SessionClaimValidator
from supertokens_python.types import MaybeAwaitable

_T = TypeVar("_T", bound=Callable[..., Any])


def verify_session(
    anti_csrf_check: Union[bool, None] = None,
    session_required: bool = True,
    override_global_claim_validators: Optional[
        Callable[
            [List[SessionClaimValidator], SessionContainer, Dict[str, Any]],
            MaybeAwaitable[List[SessionClaimValidator]],
        ]
    ] = None,
    user_context: Union[None, Dict[str, Any]] = None,
) -> Callable[[_T], _T]:
    if user_context is None:
        user_context = {}

    def session_verify(f: _T) -> _T:
        from django.http import HttpRequest

        @wraps(f)
        async def wrapped_function(request: HttpRequest, *args: Any, **kwargs: Any):
            from django.http import JsonResponse

            try:
                baseRequest = DjangoRequest(request)
                recipe = SessionRecipe.get_instance()
                session = await recipe.verify_session(
                    baseRequest,
                    anti_csrf_check,
                    session_required,
                    override_global_claim_validators,
                    user_context,
                )
                if session is None:
                    if session_required:
                        raise Exception("Should never come here")
                    baseRequest.set_session_as_none()
                else:
                    baseRequest.set_session(session)
                return await f(baseRequest.request, *args, **kwargs)
            except SuperTokensError as e:
                response = DjangoResponse(JsonResponse({}))
                result = await Supertokens.get_instance().handle_supertokens_error(
                    DjangoRequest(request), e, response
                )
                if isinstance(result, DjangoResponse):
                    return result.response
                raise Exception("Should never come here")

        return cast(_T, wrapped_function)

    return session_verify
